# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Suffix
#

import os
import itertools
import logging
import pickle
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentMap
from pyomo.common.log import LoggingIntercept
from pyomo.core.base.suffix import (
    active_export_suffix_generator,
    export_suffix_generator,
    active_import_suffix_generator,
    import_suffix_generator,
    active_local_suffix_generator,
    local_suffix_generator,
    active_suffix_generator,
    suffix_generator,
    SuffixFinder,
)
from pyomo.environ import (
    ConcreteModel,
    Suffix,
    Var,
    Param,
    Set,
    Objective,
    Constraint,
    Block,
    sum_product,
)

from io import StringIO


def simple_con_rule(model, i):
    return model.x[i] == 1


def simple_obj_rule(model, i):
    return model.x[i]


class TestSuffixMethods(unittest.TestCase):
    def test_suffix_debug(self):
        with LoggingIntercept(level=logging.DEBUG) as OUT:
            m = ConcreteModel()
            m.s = Suffix()
            m.foo = Suffix(rule=[])
        print(OUT.getvalue())
        self.assertEqual(
            OUT.getvalue(),
            "Constructing ConcreteModel 'ConcreteModel', from data=None\n"
            "Constructing Suffix 'Suffix'\n"
            "Constructing AbstractSuffix 'foo' on [Model] from data=None\n"
            "Constructing Suffix 'foo'\n"
            "Constructed component ''[Model].foo'':\n"
            "foo : Direction=LOCAL, Datatype=FLOAT\n"
            "    Key : Value\n\n",
        )

    def test_suffix_rule(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2, 3])
        m.x = Var(m.I)
        m.y = Var(m.I)
        m.c = Constraint(m.I, rule=lambda m, i: m.x[i] >= i)
        m.d = Constraint(m.I, rule=lambda m, i: m.x[i] <= -i)

        _dict = {m.c[1]: 10, m.c[2]: 20, m.c[3]: 30, m.d: 100}
        m.suffix_dict = Suffix(initialize=_dict)
        self.assertEqual(len(m.suffix_dict), 6)
        self.assertEqual(m.suffix_dict[m.c[1]], 10)
        self.assertEqual(m.suffix_dict[m.c[2]], 20)
        self.assertEqual(m.suffix_dict[m.c[3]], 30)
        self.assertEqual(m.suffix_dict[m.d[1]], 100)
        self.assertEqual(m.suffix_dict[m.d[2]], 100)
        self.assertEqual(m.suffix_dict[m.d[3]], 100)

        # check double-construction
        _dict[m.c[1]] = 1000
        m.suffix_dict.construct()
        self.assertEqual(len(m.suffix_dict), 6)
        self.assertEqual(m.suffix_dict[m.c[1]], 10)

        m.suffix_cmap = Suffix(
            initialize=ComponentMap(
                [(m.x[1], 10), (m.x[2], 20), (m.x[3], 30), (m.y, 100)]
            )
        )
        self.assertEqual(len(m.suffix_dict), 6)
        self.assertEqual(m.suffix_cmap[m.x[1]], 10)
        self.assertEqual(m.suffix_cmap[m.x[2]], 20)
        self.assertEqual(m.suffix_cmap[m.x[3]], 30)
        self.assertEqual(m.suffix_cmap[m.y[1]], 100)
        self.assertEqual(m.suffix_cmap[m.y[2]], 100)
        self.assertEqual(m.suffix_cmap[m.y[3]], 100)

        m.suffix_list = Suffix(
            initialize=[(m.x[1], 10), (m.x[2], 20), (m.x[3], 30), (m.y, 100)]
        )
        self.assertEqual(len(m.suffix_dict), 6)
        self.assertEqual(m.suffix_list[m.x[1]], 10)
        self.assertEqual(m.suffix_list[m.x[2]], 20)
        self.assertEqual(m.suffix_list[m.x[3]], 30)
        self.assertEqual(m.suffix_list[m.y[1]], 100)
        self.assertEqual(m.suffix_list[m.y[2]], 100)
        self.assertEqual(m.suffix_list[m.y[3]], 100)

        def gen_init():
            yield (m.x[1], 10)
            yield (m.x[2], 20)
            yield (m.x[3], 30)
            yield (m.y, 100)

        m.suffix_generator = Suffix(initialize=gen_init())
        self.assertEqual(len(m.suffix_dict), 6)
        self.assertEqual(m.suffix_generator[m.x[1]], 10)
        self.assertEqual(m.suffix_generator[m.x[2]], 20)
        self.assertEqual(m.suffix_generator[m.x[3]], 30)
        self.assertEqual(m.suffix_generator[m.y[1]], 100)
        self.assertEqual(m.suffix_generator[m.y[2]], 100)
        self.assertEqual(m.suffix_generator[m.y[3]], 100)

        def genfcn_init(m, i):
            yield (m.x[1], 10)
            yield (m.x[2], 20)
            yield (m.x[3], 30)
            yield (m.y, 100)

        m.suffix_generator_fcn = Suffix(initialize=genfcn_init)
        self.assertEqual(len(m.suffix_dict), 6)
        self.assertEqual(m.suffix_generator_fcn[m.x[1]], 10)
        self.assertEqual(m.suffix_generator_fcn[m.x[2]], 20)
        self.assertEqual(m.suffix_generator_fcn[m.x[3]], 30)
        self.assertEqual(m.suffix_generator_fcn[m.y[1]], 100)
        self.assertEqual(m.suffix_generator_fcn[m.y[2]], 100)
        self.assertEqual(m.suffix_generator_fcn[m.y[3]], 100)

    # test import_enabled
    def test_import_enabled(self):
        model = ConcreteModel()
        model.test_implicit = Suffix()
        self.assertFalse(model.test_implicit.import_enabled())

        model.test_local = Suffix(direction=Suffix.LOCAL)
        self.assertFalse(model.test_local.import_enabled())

        model.test_out = Suffix(direction=Suffix.IMPORT)
        self.assertTrue(model.test_out.import_enabled())

        model.test_in = Suffix(direction=Suffix.EXPORT)
        self.assertFalse(model.test_in.import_enabled())

        model.test_inout = Suffix(direction=Suffix.IMPORT_EXPORT)
        self.assertTrue(model.test_inout.import_enabled())

    # test export_enabled
    def test_export_enabled(self):
        model = ConcreteModel()

        model.test_implicit = Suffix()
        self.assertFalse(model.test_implicit.export_enabled())

        model.test_local = Suffix(direction=Suffix.LOCAL)
        self.assertFalse(model.test_local.export_enabled())

        model.test_out = Suffix(direction=Suffix.IMPORT)
        self.assertFalse(model.test_out.export_enabled())

        model.test_in = Suffix(direction=Suffix.EXPORT)
        self.assertTrue(model.test_in.export_enabled())

        model.test_inout = Suffix(direction=Suffix.IMPORT_EXPORT)
        self.assertTrue(model.test_inout.export_enabled())

    # test set_value and getValue
    # and if Var arrays are correctly expanded
    def test_set_value_getValue_Var1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3], dense=True)

        model.junk.set_value(model.X, 1.0)
        model.junk.set_value(model.X[1], 2.0)

        self.assertEqual(model.junk.get(model.X), None)
        self.assertEqual(model.junk.get(model.X[1]), 2.0)
        self.assertEqual(model.junk.get(model.X[2]), 1.0)
        self.assertEqual(model.junk.get(model.x), None)

        model.junk.set_value(model.x, 3.0)
        model.junk.set_value(model.X[2], 3.0)

        self.assertEqual(model.junk.get(model.X), None)
        self.assertEqual(model.junk.get(model.X[1]), 2.0)
        self.assertEqual(model.junk.get(model.X[2]), 3.0)
        self.assertEqual(model.junk.get(model.x), 3.0)

        model.junk.set_value(model.X, 1.0, expand=False)

        self.assertEqual(model.junk.get(model.X), 1.0)

    # test set_value and getValue
    # and if Var arrays are correctly expanded
    def test_set_value_getValue_Var2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3], dense=True)

        model.X.set_suffix_value('junk', 1.0)
        model.X[1].set_suffix_value('junk', 2.0)

        self.assertEqual(model.X.get_suffix_value('junk'), None)
        self.assertEqual(model.X[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.X[2].get_suffix_value('junk'), 1.0)
        self.assertEqual(model.x.get_suffix_value('junk'), None)

        model.x.set_suffix_value('junk', 3.0)
        model.X[2].set_suffix_value('junk', 3.0)

        self.assertEqual(model.X.get_suffix_value('junk'), None)
        self.assertEqual(model.X[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.X[2].get_suffix_value('junk'), 3.0)
        self.assertEqual(model.x.get_suffix_value('junk'), 3.0)

        model.X.set_suffix_value('junk', 1.0, expand=False)

        self.assertEqual(model.X.get_suffix_value('junk'), 1.0)

    # test set_value and getValue
    # and if Var arrays are correctly expanded
    def test_set_value_getValue_Var3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3], dense=True)

        model.X.set_suffix_value(model.junk, 1.0)
        model.X[1].set_suffix_value(model.junk, 2.0)

        self.assertEqual(model.X.get_suffix_value(model.junk), None)
        self.assertEqual(model.X[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.X[2].get_suffix_value(model.junk), 1.0)
        self.assertEqual(model.x.get_suffix_value(model.junk), None)

        model.x.set_suffix_value(model.junk, 3.0)
        model.X[2].set_suffix_value(model.junk, 3.0)

        self.assertEqual(model.X.get_suffix_value(model.junk), None)
        self.assertEqual(model.X[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.X[2].get_suffix_value(model.junk), 3.0)
        self.assertEqual(model.x.get_suffix_value(model.junk), 3.0)

        model.X.set_suffix_value(model.junk, 1.0, expand=False)

        self.assertEqual(model.X.get_suffix_value(model.junk), 1.0)

    # test set_value and getValue
    # and if Constraint arrays are correctly expanded
    def test_set_value_getValue_Constraint1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.c = Constraint(expr=model.x >= 1)
        model.C = Constraint([1, 2, 3], rule=lambda model, i: model.X[i] >= 1)

        model.junk.set_value(model.C, 1.0)
        model.junk.set_value(model.C[1], 2.0)

        self.assertEqual(model.junk.get(model.C), None)
        self.assertEqual(model.junk.get(model.C[1]), 2.0)
        self.assertEqual(model.junk.get(model.C[2]), 1.0)

        self.assertEqual(model.junk.get(model.c), None)

        model.junk.set_value(model.c, 3.0)
        model.junk.set_value(model.C[2], 3.0)

        self.assertEqual(model.junk.get(model.C), None)
        self.assertEqual(model.junk.get(model.C[1]), 2.0)
        self.assertEqual(model.junk.get(model.C[2]), 3.0)
        self.assertEqual(model.junk.get(model.c), 3.0)

        model.junk.set_value(model.C, 1.0, expand=False)

        self.assertEqual(model.junk.get(model.C), 1.0)

    # test set_value and getValue
    # and if Constraint arrays are correctly expanded
    def test_set_value_getValue_Constraint2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.c = Constraint(expr=model.x >= 1)
        model.C = Constraint([1, 2, 3], rule=lambda model, i: model.X[i] >= 1)

        model.C.set_suffix_value('junk', 1.0)
        model.C[1].set_suffix_value('junk', 2.0)

        self.assertEqual(model.C.get_suffix_value('junk'), None)
        self.assertEqual(model.C[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.C[2].get_suffix_value('junk'), 1.0)
        self.assertEqual(model.c.get_suffix_value('junk'), None)

        model.c.set_suffix_value('junk', 3.0)
        model.C[2].set_suffix_value('junk', 3.0)

        self.assertEqual(model.C.get_suffix_value('junk'), None)
        self.assertEqual(model.C[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.C[2].get_suffix_value('junk'), 3.0)
        self.assertEqual(model.c.get_suffix_value('junk'), 3.0)

        model.C.set_suffix_value('junk', 1.0, expand=False)

        self.assertEqual(model.C.get_suffix_value('junk'), 1.0)

    # test set_value and getValue
    # and if Constraint arrays are correctly expanded
    def test_set_value_getValue_Constraint3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.c = Constraint(expr=model.x >= 1)
        model.C = Constraint([1, 2, 3], rule=lambda model, i: model.X[i] >= 1)

        model.C.set_suffix_value(model.junk, 1.0)
        model.C[1].set_suffix_value(model.junk, 2.0)

        self.assertEqual(model.C.get_suffix_value(model.junk), None)
        self.assertEqual(model.C[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.C[2].get_suffix_value(model.junk), 1.0)
        self.assertEqual(model.c.get_suffix_value(model.junk), None)

        model.c.set_suffix_value(model.junk, 3.0)
        model.C[2].set_suffix_value(model.junk, 3.0)

        self.assertEqual(model.C.get_suffix_value(model.junk), None)
        self.assertEqual(model.C[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.C[2].get_suffix_value(model.junk), 3.0)
        self.assertEqual(model.c.get_suffix_value(model.junk), 3.0)

        model.C.set_suffix_value(model.junk, 1.0, expand=False)

        self.assertEqual(model.C.get_suffix_value(model.junk), 1.0)

    # test set_value and getValue
    # and if Objective arrays are correctly expanded
    def test_set_value_getValue_Objective1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.obj = Objective(expr=sum_product(model.X) + model.x)
        model.OBJ = Objective([1, 2, 3], rule=lambda model, i: model.X[i])

        model.junk.set_value(model.OBJ, 1.0)
        model.junk.set_value(model.OBJ[1], 2.0)

        self.assertEqual(model.junk.get(model.OBJ), None)
        self.assertEqual(model.junk.get(model.OBJ[1]), 2.0)
        self.assertEqual(model.junk.get(model.OBJ[2]), 1.0)
        self.assertEqual(model.junk.get(model.obj), None)

        model.junk.set_value(model.obj, 3.0)
        model.junk.set_value(model.OBJ[2], 3.0)

        self.assertEqual(model.junk.get(model.OBJ), None)
        self.assertEqual(model.junk.get(model.OBJ[1]), 2.0)
        self.assertEqual(model.junk.get(model.OBJ[2]), 3.0)
        self.assertEqual(model.junk.get(model.obj), 3.0)

        model.junk.set_value(model.OBJ, 1.0, expand=False)

        self.assertEqual(model.junk.get(model.OBJ), 1.0)

    # test set_value and getValue
    # and if Objective arrays are correctly expanded
    def test_set_value_getValue_Objective2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.obj = Objective(expr=sum_product(model.X) + model.x)
        model.OBJ = Objective([1, 2, 3], rule=lambda model, i: model.X[i])

        model.OBJ.set_suffix_value('junk', 1.0)
        model.OBJ[1].set_suffix_value('junk', 2.0)

        self.assertEqual(model.OBJ.get_suffix_value('junk'), None)
        self.assertEqual(model.OBJ[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.OBJ[2].get_suffix_value('junk'), 1.0)
        self.assertEqual(model.obj.get_suffix_value('junk'), None)

        model.obj.set_suffix_value('junk', 3.0)
        model.OBJ[2].set_suffix_value('junk', 3.0)

        self.assertEqual(model.OBJ.get_suffix_value('junk'), None)
        self.assertEqual(model.OBJ[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.OBJ[2].get_suffix_value('junk'), 3.0)
        self.assertEqual(model.obj.get_suffix_value('junk'), 3.0)

        model.OBJ.set_suffix_value('junk', 1.0, expand=False)

        self.assertEqual(model.OBJ.get_suffix_value('junk'), 1.0)

    # test set_value and getValue
    # and if Objective arrays are correctly expanded
    def test_set_value_getValue_Objective3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.obj = Objective(expr=sum_product(model.X) + model.x)
        model.OBJ = Objective([1, 2, 3], rule=lambda model, i: model.X[i])

        model.OBJ.set_suffix_value(model.junk, 1.0)
        model.OBJ[1].set_suffix_value(model.junk, 2.0)

        self.assertEqual(model.OBJ.get_suffix_value(model.junk), None)
        self.assertEqual(model.OBJ[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.OBJ[2].get_suffix_value(model.junk), 1.0)
        self.assertEqual(model.obj.get_suffix_value(model.junk), None)

        model.obj.set_suffix_value(model.junk, 3.0)
        model.OBJ[2].set_suffix_value(model.junk, 3.0)

        self.assertEqual(model.OBJ.get_suffix_value(model.junk), None)
        self.assertEqual(model.OBJ[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.OBJ[2].get_suffix_value(model.junk), 3.0)
        self.assertEqual(model.obj.get_suffix_value(model.junk), 3.0)

        model.OBJ.set_suffix_value(model.junk, 1.0, expand=False)

        self.assertEqual(model.OBJ.get_suffix_value(model.junk), 1.0)

    # test set_value and getValue
    # and if mutable Param arrays are correctly expanded
    def test_set_value_getValue_mutableParam1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.p = Param(initialize=1.0, mutable=True)
        model.P = Param([1, 2, 3], initialize=1.0, mutable=True)

        model.junk.set_value(model.P, 1.0)
        model.junk.set_value(model.P[1], 2.0)

        self.assertEqual(model.junk.get(model.P), None)
        self.assertEqual(model.junk.get(model.P[1]), 2.0)
        self.assertEqual(model.junk.get(model.P[2]), 1.0)
        self.assertEqual(model.junk.get(model.p), None)

        model.junk.set_value(model.p, 3.0)
        model.junk.set_value(model.P[2], 3.0)

        self.assertEqual(model.junk.get(model.P), None)
        self.assertEqual(model.junk.get(model.P[1]), 2.0)
        self.assertEqual(model.junk.get(model.P[2]), 3.0)
        self.assertEqual(model.junk.get(model.p), 3.0)

        model.junk.set_value(model.P, 1.0, expand=False)

        self.assertEqual(model.junk.get(model.P), 1.0)

    # test set_value and getValue
    # and if mutable Param arrays are correctly expanded
    def test_set_value_getValue_mutableParam2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.p = Param(initialize=1.0, mutable=True)
        model.P = Param([1, 2, 3], initialize=1.0, mutable=True)

        model.P.set_suffix_value('junk', 1.0)
        model.P[1].set_suffix_value('junk', 2.0)

        self.assertEqual(model.P.get_suffix_value('junk'), None)
        self.assertEqual(model.P[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.P[2].get_suffix_value('junk'), 1.0)
        self.assertEqual(model.p.get_suffix_value('junk'), None)

        model.p.set_suffix_value('junk', 3.0)
        model.P[2].set_suffix_value('junk', 3.0)

        self.assertEqual(model.P.get_suffix_value('junk'), None)
        self.assertEqual(model.P[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.P[2].get_suffix_value('junk'), 3.0)
        self.assertEqual(model.p.get_suffix_value('junk'), 3.0)

        model.P.set_suffix_value('junk', 1.0, expand=False)

        self.assertEqual(model.P.get_suffix_value('junk'), 1.0)

    # test set_value and getValue
    # and if mutable Param arrays are correctly expanded
    def test_set_value_getValue_mutableParam3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.p = Param(initialize=1.0, mutable=True)
        model.P = Param([1, 2, 3], initialize=1.0, mutable=True)

        model.P.set_suffix_value(model.junk, 1.0)
        model.P[1].set_suffix_value(model.junk, 2.0)

        self.assertEqual(model.P.get_suffix_value(model.junk), None)
        self.assertEqual(model.P[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.P[2].get_suffix_value(model.junk), 1.0)
        self.assertEqual(model.p.get_suffix_value(model.junk), None)

        model.p.set_suffix_value(model.junk, 3.0)
        model.P[2].set_suffix_value(model.junk, 3.0)

        self.assertEqual(model.P.get_suffix_value(model.junk), None)
        self.assertEqual(model.P[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.P[2].get_suffix_value(model.junk), 3.0)
        self.assertEqual(model.p.get_suffix_value(model.junk), 3.0)

        model.P.set_suffix_value(model.junk, 1.0, expand=False)

        self.assertEqual(model.P.get_suffix_value(model.junk), 1.0)

    # test set_value and getValue
    # and if immutable Param arrays are correctly expanded
    def test_set_value_getValue_immutableParam1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.p = Param(initialize=1.0, mutable=False)
        model.P = Param([1, 2, 3], initialize=1.0, mutable=False)

        self.assertEqual(model.junk.get(model.P), None)

        model.junk.set_value(model.P, 1.0, expand=False)

        self.assertEqual(model.junk.get(model.P), 1.0)

    # test set_value and getValue
    # and if immutable Param arrays are correctly expanded
    def test_set_value_getValue_immutableParam2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.p = Param(initialize=1.0, mutable=False)
        model.P = Param([1, 2, 3], initialize=1.0, mutable=False)

        self.assertEqual(model.P.get_suffix_value('junk'), None)

        model.P.set_suffix_value('junk', 1.0, expand=False)

        self.assertEqual(model.P.get_suffix_value('junk'), 1.0)

    # test set_value and getValue
    # and if immutable Param arrays are correctly expanded
    def test_set_value_getValue_immutableParam3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.p = Param(initialize=1.0, mutable=False)
        model.P = Param([1, 2, 3], initialize=1.0, mutable=False)

        self.assertEqual(model.P.get_suffix_value(model.junk), None)

        model.P.set_suffix_value(model.junk, 1.0, expand=False)

        self.assertEqual(model.P.get_suffix_value(model.junk), 1.0)

    # test set_value and getValue
    # and if Set arrays are correctly expanded
    def test_set_value_getValue_Set1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.s = Set(initialize=[1, 2, 3])
        model.S = Set([1, 2, 3], initialize={1: [1, 2, 3], 2: [1, 2, 3], 3: [1, 2, 3]})

        model.junk.set_value(model.S, 1.0)
        model.junk.set_value(model.S[1], 2.0)

        self.assertEqual(model.junk.get(model.S), None)
        self.assertEqual(model.junk.get(model.S[1]), 2.0)
        self.assertEqual(model.junk.get(model.S[2]), 1.0)
        self.assertEqual(model.junk.get(model.s), None)

        model.junk.set_value(model.s, 3.0)
        model.junk.set_value(model.S[2], 3.0)

        self.assertEqual(model.junk.get(model.S), None)
        self.assertEqual(model.junk.get(model.S[1]), 2.0)
        self.assertEqual(model.junk.get(model.S[2]), 3.0)
        self.assertEqual(model.junk.get(model.s), 3.0)

        model.junk.set_value(model.S, 1.0, expand=False)

        self.assertEqual(model.junk.get(model.S), 1.0)

    # test set_value and getValue
    # and if Set arrays are correctly expanded
    def test_set_value_getValue_Set2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.s = Set(initialize=[1, 2, 3])
        model.S = Set([1, 2, 3], initialize={1: [1, 2, 3], 2: [1, 2, 3], 3: [1, 2, 3]})

        model.S.set_suffix_value('junk', 1.0)
        model.S[1].set_suffix_value('junk', 2.0)

        self.assertEqual(model.S.get_suffix_value('junk'), None)
        self.assertEqual(model.S[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.S[2].get_suffix_value('junk'), 1.0)
        self.assertEqual(model.s.get_suffix_value('junk'), None)

        model.s.set_suffix_value('junk', 3.0)
        model.S[2].set_suffix_value('junk', 3.0)

        self.assertEqual(model.S.get_suffix_value('junk'), None)
        self.assertEqual(model.S[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.S[2].get_suffix_value('junk'), 3.0)
        self.assertEqual(model.s.get_suffix_value('junk'), 3.0)

        model.S.set_suffix_value('junk', 1.0, expand=False)

        self.assertEqual(model.S.get_suffix_value('junk'), 1.0)

    # test set_value and getValue
    # and if Set arrays are correctly expanded
    def test_set_value_getValue_Set3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1, 2, 3])
        model.s = Set(initialize=[1, 2, 3])
        model.S = Set([1, 2, 3], initialize={1: [1, 2, 3], 2: [1, 2, 3], 3: [1, 2, 3]})

        model.S.set_suffix_value(model.junk, 1.0)
        model.S[1].set_suffix_value(model.junk, 2.0)

        self.assertEqual(model.S.get_suffix_value(model.junk), None)
        self.assertEqual(model.S[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.S[2].get_suffix_value(model.junk), 1.0)
        self.assertEqual(model.s.get_suffix_value(model.junk), None)

        model.s.set_suffix_value(model.junk, 3.0)
        model.S[2].set_suffix_value(model.junk, 3.0)

        self.assertEqual(model.S.get_suffix_value(model.junk), None)
        self.assertEqual(model.S[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.S[2].get_suffix_value(model.junk), 3.0)
        self.assertEqual(model.s.get_suffix_value(model.junk), 3.0)

        model.S.set_suffix_value(model.junk, 1.0, expand=False)

        self.assertEqual(model.S.get_suffix_value(model.junk), 1.0)

    # test set_value and getValue
    # and if Block arrays are correctly expanded
    def test_set_value_getValue_Block1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.b = Block()
        model.B = Block([1, 2, 3])

        # make sure each BlockData gets constructed
        model.B[1].x = 1
        model.B[2].x = 2
        model.B[3].x = 3

        model.junk.set_value(model.B, 1.0)
        model.junk.set_value(model.B[1], 2.0)

        self.assertEqual(model.junk.get(model.B), None)
        self.assertEqual(model.junk.get(model.B[1]), 2.0)
        self.assertEqual(model.junk.get(model.B[2]), 1.0)
        self.assertEqual(model.junk.get(model.b), None)

        model.junk.set_value(model.b, 3.0)
        model.junk.set_value(model.B[2], 3.0)

        self.assertEqual(model.junk.get(model.B), None)
        self.assertEqual(model.junk.get(model.B[1]), 2.0)
        self.assertEqual(model.junk.get(model.B[2]), 3.0)
        self.assertEqual(model.junk.get(model.b), 3.0)

        model.junk.set_value(model.B, 1.0, expand=False)

        self.assertEqual(model.junk.get(model.B), 1.0)

    # test set_value and getValue
    # and if Block arrays are correctly expanded
    def test_set_value_getValue_Block2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.b = Block()
        model.B = Block([1, 2, 3])

        # make sure each BlockData gets constructed
        model.B[1].x = 1
        model.B[2].x = 2
        model.B[3].x = 3

        model.B.set_suffix_value('junk', 1.0)
        model.B[1].set_suffix_value('junk', 2.0)

        self.assertEqual(model.B.get_suffix_value('junk'), None)
        self.assertEqual(model.B[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.B[2].get_suffix_value('junk'), 1.0)
        self.assertEqual(model.b.get_suffix_value('junk'), None)

        model.b.set_suffix_value('junk', 3.0)
        model.B[2].set_suffix_value('junk', 3.0)

        self.assertEqual(model.B.get_suffix_value('junk'), None)
        self.assertEqual(model.B[1].get_suffix_value('junk'), 2.0)
        self.assertEqual(model.B[2].get_suffix_value('junk'), 3.0)
        self.assertEqual(model.b.get_suffix_value('junk'), 3.0)

        model.B.set_suffix_value('junk', 1.0, expand=False)

        self.assertEqual(model.B.get_suffix_value('junk'), 1.0)

    # test set_value and getValue
    # and if Block arrays are correctly expanded
    def test_set_value_getValue_Block3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.b = Block()
        model.B = Block([1, 2, 3])

        # make sure each BlockData gets constructed
        model.B[1].x = 1
        model.B[2].x = 2
        model.B[3].x = 3

        model.B.set_suffix_value(model.junk, 1.0)
        model.B[1].set_suffix_value(model.junk, 2.0)

        self.assertEqual(model.B.get_suffix_value(model.junk), None)
        self.assertEqual(model.B[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.B[2].get_suffix_value(model.junk), 1.0)
        self.assertEqual(model.b.get_suffix_value(model.junk), None)

        model.b.set_suffix_value(model.junk, 3.0)
        model.B[2].set_suffix_value(model.junk, 3.0)

        self.assertEqual(model.B.get_suffix_value(model.junk), None)
        self.assertEqual(model.B[1].get_suffix_value(model.junk), 2.0)
        self.assertEqual(model.B[2].get_suffix_value(model.junk), 3.0)
        self.assertEqual(model.b.get_suffix_value(model.junk), 3.0)

        model.B.set_suffix_value(model.junk, 1.0, expand=False)

        self.assertEqual(model.B.get_suffix_value(model.junk), 1.0)

    # test set_value with no component argument
    def test_set_all_values1(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1, 2, 3], dense=True)
        model.z = Var([1, 2, 3], dense=True)

        model.junk.set_value(model.y[2], 1.0)
        model.junk.set_value(model.z, 2.0)

        self.assertTrue(model.junk.get(model.x) is None)
        self.assertTrue(model.junk.get(model.y) is None)
        self.assertTrue(model.junk.get(model.y[1]) is None)
        self.assertEqual(model.junk.get(model.y[2]), 1.0)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[1]), 2.0)

        model.junk.set_all_values(3.0)

        self.assertTrue(model.junk.get(model.x) is None)
        self.assertTrue(model.junk.get(model.y) is None)
        self.assertTrue(model.junk.get(model.y[1]) is None)
        self.assertEqual(model.junk.get(model.y[2]), 3.0)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[1]), 3.0)

    # test set_value with no component argument
    def test_set_all_values2(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1, 2, 3], dense=True)
        model.z = Var([1, 2, 3], dense=True)

        model.y[2].set_suffix_value('junk', 1.0)
        model.z.set_suffix_value('junk', 2.0)

        self.assertTrue(model.x.get_suffix_value('junk') is None)
        self.assertTrue(model.y.get_suffix_value('junk') is None)
        self.assertTrue(model.y[1].get_suffix_value('junk') is None)
        self.assertEqual(model.y[2].get_suffix_value('junk'), 1.0)
        self.assertEqual(model.z.get_suffix_value('junk'), None)
        self.assertEqual(model.z[1].get_suffix_value('junk'), 2.0)

        model.junk.set_all_values(3.0)

        self.assertTrue(model.x.get_suffix_value('junk') is None)
        self.assertTrue(model.y.get_suffix_value('junk') is None)
        self.assertTrue(model.y[1].get_suffix_value('junk') is None)
        self.assertEqual(model.y[2].get_suffix_value('junk'), 3.0)
        self.assertEqual(model.z.get_suffix_value('junk'), None)
        self.assertEqual(model.z[1].get_suffix_value('junk'), 3.0)

    # test set_value with no component argument
    def test_set_all_values3(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1, 2, 3], dense=True)
        model.z = Var([1, 2, 3], dense=True)

        model.y[2].set_suffix_value(model.junk, 1.0)
        model.z.set_suffix_value(model.junk, 2.0)

        self.assertTrue(model.x.get_suffix_value(model.junk) is None)
        self.assertTrue(model.y.get_suffix_value(model.junk) is None)
        self.assertTrue(model.y[1].get_suffix_value(model.junk) is None)
        self.assertEqual(model.y[2].get_suffix_value(model.junk), 1.0)
        self.assertEqual(model.z.get_suffix_value(model.junk), None)
        self.assertEqual(model.z[1].get_suffix_value(model.junk), 2.0)

        model.junk.set_all_values(3.0)

        self.assertTrue(model.x.get_suffix_value(model.junk) is None)
        self.assertTrue(model.y.get_suffix_value(model.junk) is None)
        self.assertTrue(model.y[1].get_suffix_value(model.junk) is None)
        self.assertEqual(model.y[2].get_suffix_value(model.junk), 3.0)
        self.assertEqual(model.z.get_suffix_value(model.junk), None)
        self.assertEqual(model.z[1].get_suffix_value(model.junk), 3.0)

    # test update_values
    def test_update_values(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var()
        model.z = Var([1, 2])
        model.junk.set_value(model.x, 0.0)
        self.assertEqual(model.junk.get(model.x), 0.0)
        self.assertEqual(model.junk.get(model.y), None)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[1]), None)
        self.assertEqual(model.junk.get(model.z[2]), None)
        model.junk.update_values([(model.x, 1.0), (model.y, 2.0), (model.z, 3.0)])
        self.assertEqual(model.junk.get(model.x), 1.0)
        self.assertEqual(model.junk.get(model.y), 2.0)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[1]), 3.0)
        self.assertEqual(model.junk.get(model.z[2]), 3.0)
        model.junk.clear()
        model.junk.update_values(
            [(model.x, 1.0), (model.y, 2.0), (model.z, 3.0)], expand=False
        )
        self.assertEqual(model.junk.get(model.x), 1.0)
        self.assertEqual(model.junk.get(model.y), 2.0)
        self.assertEqual(model.junk.get(model.z), 3.0)
        self.assertEqual(model.junk.get(model.z[1]), None)
        self.assertEqual(model.junk.get(model.z[2]), None)

    # test clear_value
    def test_clear_value(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1, 2, 3], dense=True)
        model.z = Var([1, 2, 3], dense=True)

        model.junk.set_value(model.x, -1.0)
        model.junk.set_value(model.y, -2.0)
        model.junk.set_value(model.y[2], 1.0)
        model.junk.set_value(model.z, 2.0)
        model.junk.set_value(model.z[1], 4.0)

        self.assertEqual(model.junk.get(model.x), -1.0)
        self.assertEqual(model.junk.get(model.y), None)
        self.assertEqual(model.junk.get(model.y[1]), -2.0)
        self.assertEqual(model.junk.get(model.y[2]), 1.0)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[2]), 2.0)
        self.assertEqual(model.junk.get(model.z[1]), 4.0)

        model.junk.clear_value(model.y)

        self.assertEqual(model.junk.get(model.x), -1.0)
        self.assertEqual(model.junk.get(model.y), None)
        self.assertEqual(model.junk.get(model.y[1]), None)
        self.assertEqual(model.junk.get(model.y[2]), None)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[2]), 2.0)
        self.assertEqual(model.junk.get(model.z[1]), 4.0)

        model.junk.clear_value(model.x)

        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.y), None)
        self.assertEqual(model.junk.get(model.y[1]), None)
        self.assertEqual(model.junk.get(model.y[2]), None)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[2]), 2.0)
        self.assertEqual(model.junk.get(model.z[1]), 4.0)

        # Clearing a scalar that is not there does not raise an error
        model.junk.clear_value(model.x)

        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.y), None)
        self.assertEqual(model.junk.get(model.y[1]), None)
        self.assertEqual(model.junk.get(model.y[2]), None)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[2]), 2.0)
        self.assertEqual(model.junk.get(model.z[1]), 4.0)

        model.junk.clear_value(model.z[1])

        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.y), None)
        self.assertEqual(model.junk.get(model.y[1]), None)
        self.assertEqual(model.junk.get(model.y[2]), None)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[2]), 2.0)
        self.assertEqual(model.junk.get(model.z[1]), None)

        # Clearing an indexed component with missing indices does not raise an error
        model.junk.clear_value(model.z)

        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.y), None)
        self.assertEqual(model.junk.get(model.y[1]), None)
        self.assertEqual(model.junk.get(model.y[2]), None)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[2]), None)
        self.assertEqual(model.junk.get(model.z[1]), None)

    # test clear_value no args
    def test_clear_all_values(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1, 2, 3], dense=True)
        model.z = Var([1, 2, 3], dense=True)

        model.junk.set_value(model.y[2], 1.0)
        model.junk.set_value(model.z, 2.0)

        self.assertTrue(model.junk.get(model.x) is None)
        self.assertTrue(model.junk.get(model.y) is None)
        self.assertTrue(model.junk.get(model.y[1]) is None)
        self.assertEqual(model.junk.get(model.y[2]), 1.0)
        self.assertEqual(model.junk.get(model.z), None)
        self.assertEqual(model.junk.get(model.z[1]), 2.0)

        model.junk.clear_all_values()

        self.assertTrue(model.junk.get(model.x) is None)
        self.assertTrue(model.junk.get(model.y) is None)
        self.assertTrue(model.junk.get(model.y[1]) is None)
        self.assertTrue(model.junk.get(model.y[2]) is None)
        self.assertTrue(model.junk.get(model.z) is None)
        self.assertTrue(model.junk.get(model.z[1]) is None)

    # test set_datatype and get_datatype
    def test_set_datatype_get_datatype(self):
        model = ConcreteModel()
        model.junk = Suffix(datatype=Suffix.FLOAT)
        self.assertEqual(model.junk.datatype, Suffix.FLOAT)
        model.junk.datatype = Suffix.INT
        self.assertEqual(model.junk.datatype, Suffix.INT)
        model.junk.datatype = None
        self.assertEqual(model.junk.datatype, None)
        model.junk.datatype = 'FLOAT'
        self.assertEqual(model.junk.datatype, Suffix.FLOAT)
        model.junk.datatype = 'INT'
        self.assertEqual(model.junk.datatype, Suffix.INT)
        model.junk.datatype = 4
        self.assertEqual(model.junk.datatype, Suffix.FLOAT)
        model.junk.datatype = 0
        self.assertEqual(model.junk.datatype, Suffix.INT)

        with LoggingIntercept() as LOG:
            model.junk.set_datatype(None)
        self.assertEqual(model.junk.datatype, None)
        self.assertRegex(
            LOG.getvalue().replace("\n", " "),
            "^DEPRECATED: Suffix.set_datatype is replaced with the "
            "Suffix.datatype property",
        )

        model.junk.datatype = 'FLOAT'
        with LoggingIntercept() as LOG:
            self.assertEqual(model.junk.get_datatype(), Suffix.FLOAT)
        self.assertRegex(
            LOG.getvalue().replace("\n", " "),
            "^DEPRECATED: Suffix.get_datatype is replaced with the "
            "Suffix.datatype property",
        )

        with self.assertRaisesRegex(ValueError, "1.0 is not a valid SuffixDataType"):
            model.junk.datatype = 1.0

    # test set_direction and get_direction
    def test_set_direction_get_direction(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.LOCAL)
        self.assertEqual(model.junk.direction, Suffix.LOCAL)
        model.junk.direction = Suffix.EXPORT
        self.assertEqual(model.junk.direction, Suffix.EXPORT)
        model.junk.direction = Suffix.IMPORT
        self.assertEqual(model.junk.direction, Suffix.IMPORT)
        model.junk.direction = Suffix.IMPORT_EXPORT
        self.assertEqual(model.junk.direction, Suffix.IMPORT_EXPORT)

        with LoggingIntercept() as LOG:
            model.junk.set_direction(1)
        self.assertEqual(model.junk.direction, Suffix.EXPORT)
        self.assertRegex(
            LOG.getvalue().replace("\n", " "),
            "^DEPRECATED: Suffix.set_direction is replaced with the "
            "Suffix.direction property",
        )

        model.junk.direction = 'IMPORT'
        with LoggingIntercept() as LOG:
            self.assertEqual(model.junk.get_direction(), Suffix.IMPORT)
        self.assertRegex(
            LOG.getvalue().replace("\n", " "),
            "^DEPRECATED: Suffix.get_direction is replaced with the "
            "Suffix.direction property",
        )

        with self.assertRaisesRegex(ValueError, "'a' is not a valid SuffixDirection"):
            model.junk.direction = 'a'
        # None is allowed for datatype, but not direction
        with self.assertRaisesRegex(ValueError, "None is not a valid SuffixDirection"):
            model.junk.direction = None

    # test __str__
    def test_str(self):
        model = ConcreteModel()
        model.junk = Suffix()
        self.assertEqual(model.junk.__str__(), "junk")

    # test pprint()
    def test_pprint(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.EXPORT)
        output = StringIO()
        model.junk.pprint(ostream=output)
        self.assertEqual(
            output.getvalue(),
            "junk : Direction=EXPORT, Datatype=FLOAT\n    Key : Value\n",
        )
        model.junk.direction = Suffix.IMPORT
        output = StringIO()
        model.junk.pprint(ostream=output)
        self.assertEqual(
            output.getvalue(),
            "junk : Direction=IMPORT, Datatype=FLOAT\n    Key : Value\n",
        )
        model.junk.direction = Suffix.LOCAL
        model.junk.datatype = None
        output = StringIO()
        model.junk.pprint(ostream=output)
        self.assertEqual(
            output.getvalue(),
            "junk : Direction=LOCAL, Datatype=None\n    Key : Value\n",
        )
        model.junk.direction = Suffix.IMPORT_EXPORT
        model.junk.datatype = Suffix.INT
        output = StringIO()
        model.junk.pprint(ostream=output)
        self.assertEqual(
            output.getvalue(),
            "junk : Direction=IMPORT_EXPORT, Datatype=INT\n    Key : Value\n",
        )
        output = StringIO()
        model.pprint(ostream=output)
        self.assertEqual(
            output.getvalue(),
            """1 Suffix Declarations
    junk : Direction=IMPORT_EXPORT, Datatype=INT
        Key : Value

1 Declarations: junk
""",
        )

    # test pprint(verbose=True)
    def test_pprint_verbose(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.s = Block()
        model.s.b = Block()
        model.s.B = Block([1, 2, 3])

        model.junk.set_value(model.s.B, 1.0)
        model.junk.set_value(model.s.B[1], 2.0)

        model.junk.set_value(model.s.b, 3.0)
        model.junk.set_value(model.s.B[2], 3.0)

        output = StringIO()
        model.junk.pprint(ostream=output, verbose=True)
        self.assertEqual(
            output.getvalue(),
            """junk : Direction=LOCAL, Datatype=FLOAT
    Key    : Value
    s.B[1] :   2.0
    s.B[2] :   3.0
    s.B[3] :   1.0
       s.b :   3.0
""",
        )

    def test_active_export_suffix_generator(self):
        model = ConcreteModel()
        model.junk_EXPORT_int = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)
        model.junk_EXPORT_float = Suffix(direction=Suffix.EXPORT, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(
            direction=Suffix.IMPORT_EXPORT, datatype=Suffix.FLOAT
        )
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT, datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL, datatype=None)

        suffixes = dict(active_export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(active_export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_EXPORT_float.activate()

        suffixes = dict(active_export_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(active_export_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_export_suffix_generator(self):
        model = ConcreteModel()
        model.junk_EXPORT_int = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)
        model.junk_EXPORT_float = Suffix(direction=Suffix.EXPORT, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(
            direction=Suffix.IMPORT_EXPORT, datatype=Suffix.FLOAT
        )
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT, datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL, datatype=None)

        suffixes = dict(export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_EXPORT_float.activate()

        suffixes = dict(export_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(export_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_active_import_suffix_generator(self):
        model = ConcreteModel()
        model.junk_IMPORT_int = Suffix(direction=Suffix.IMPORT, datatype=Suffix.INT)
        model.junk_IMPORT_float = Suffix(direction=Suffix.IMPORT, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(
            direction=Suffix.IMPORT_EXPORT, datatype=Suffix.FLOAT
        )
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT, datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL, datatype=None)

        suffixes = dict(active_import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(active_import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_IMPORT_float.activate()

        suffixes = dict(active_import_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(active_import_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_import_suffix_generator(self):
        model = ConcreteModel()
        model.junk_IMPORT_int = Suffix(direction=Suffix.IMPORT, datatype=Suffix.INT)
        model.junk_IMPORT_float = Suffix(direction=Suffix.IMPORT, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(
            direction=Suffix.IMPORT_EXPORT, datatype=Suffix.FLOAT
        )
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT, datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL, datatype=None)

        suffixes = dict(import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_IMPORT_float.activate()

        suffixes = dict(import_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(import_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_active_local_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL, datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT, datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT, datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT, datatype=None)

        suffixes = dict(active_local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(active_local_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_local_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

    def test_local_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL, datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT, datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT, datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT, datatype=None)

        suffixes = dict(local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(local_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(local_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

    def test_active_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL, datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT, datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT, datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT, datatype=None)

        suffixes = dict(active_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(active_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

    def test_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL, datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL, datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT, datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT, datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT, datatype=None)

        suffixes = dict(suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(suffix_generator(model, datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)


class TestSuffixCloneUsage(unittest.TestCase):
    def test_clone_VarElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.x), None)
        model.junk.set_value(model.x, 1.0)
        self.assertEqual(model.junk.get(model.x), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.x), None)
        self.assertEqual(inst.junk.get(inst.x), 1.0)

    def test_clone_VarArray(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.x[1]), None)
        model.junk.set_value(model.x, 1.0)
        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.x[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.x[1]), None)
        self.assertEqual(inst.junk.get(inst.x[1]), 1.0)

    def test_clone_VarData(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.x[1]), None)
        model.junk.set_value(model.x[1], 1.0)
        self.assertEqual(model.junk.get(model.x[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.x[1]), None)
        self.assertEqual(inst.junk.get(inst.x[1]), 1.0)

    def test_clone_ConstraintElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=model.x == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.c), None)
        model.junk.set_value(model.c, 1.0)
        self.assertEqual(model.junk.get(model.c), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.c), None)
        self.assertEqual(inst.junk.get(inst.c), 1.0)

    def test_clone_ConstraintArray(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.c = Constraint([1, 2, 3], rule=lambda model, i: model.x[i] == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.c), None)
        self.assertEqual(model.junk.get(model.c[1]), None)
        model.junk.set_value(model.c, 1.0)
        self.assertEqual(model.junk.get(model.c), None)
        self.assertEqual(model.junk.get(model.c[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.c[1]), None)
        self.assertEqual(inst.junk.get(inst.c[1]), 1.0)

    def test_clone_ConstraintData(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.c = Constraint([1, 2, 3], rule=lambda model, i: model.x[i] == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.c[1]), None)
        model.junk.set_value(model.c[1], 1.0)
        self.assertEqual(model.junk.get(model.c[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.c[1]), None)
        self.assertEqual(inst.junk.get(inst.c[1]), 1.0)

    def test_clone_ObjectiveElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.obj = Objective(expr=model.x)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.obj), None)
        model.junk.set_value(model.obj, 1.0)
        self.assertEqual(model.junk.get(model.obj), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.obj), None)
        self.assertEqual(inst.junk.get(inst.obj), 1.0)

    def test_clone_ObjectiveArray(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.obj = Objective([1, 2, 3], rule=lambda model, i: model.x[i])
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.obj), None)
        self.assertEqual(model.junk.get(model.obj[1]), None)
        model.junk.set_value(model.obj, 1.0)
        self.assertEqual(model.junk.get(model.obj), None)
        self.assertEqual(model.junk.get(model.obj[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.obj[1]), None)
        self.assertEqual(inst.junk.get(inst.obj[1]), 1.0)

    def test_clone_ObjectiveData(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.obj = Objective([1, 2, 3], rule=lambda model, i: model.x[i])
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.obj[1]), None)
        model.junk.set_value(model.obj[1], 1.0)
        self.assertEqual(model.junk.get(model.obj[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.obj[1]), None)
        self.assertEqual(inst.junk.get(inst.obj[1]), 1.0)

    def test_clone_SimpleBlock(self):
        model = ConcreteModel()
        model.b = Block()
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.b), None)
        model.junk.set_value(model.b, 1.0)
        self.assertEqual(model.junk.get(model.b), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.b), None)
        self.assertEqual(inst.junk.get(inst.b), 1.0)

    def test_clone_IndexedBlock(self):
        model = ConcreteModel()
        model.b = Block([1, 2, 3])
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.b), None)
        self.assertEqual(model.junk.get(model.b[1]), None)
        model.junk.set_value(model.b, 1.0)
        self.assertEqual(model.junk.get(model.b), None)
        self.assertEqual(model.junk.get(model.b[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.b[1]), None)
        self.assertEqual(inst.junk.get(inst.b[1]), 1.0)

    def test_clone_BlockData(self):
        model = ConcreteModel()
        model.b = Block([1, 2, 3])
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.b[1]), None)
        model.junk.set_value(model.b[1], 1.0)
        self.assertEqual(model.junk.get(model.b[1]), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model.b[1]), None)
        self.assertEqual(inst.junk.get(inst.b[1]), 1.0)

    def test_clone_model(self):
        model = ConcreteModel()
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model), None)
        model.junk.set_value(model, 1.0)
        self.assertEqual(model.junk.get(model), 1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.get(model), None)
        self.assertEqual(inst.junk.get(inst), 1.0)


class TestSuffixPickleUsage(unittest.TestCase):
    def test_pickle_VarElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.x), None)
        model.junk.set_value(model.x, 1.0)
        self.assertEqual(model.junk.get(model.x), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.x), None)
        self.assertEqual(inst.junk.get(inst.x), 1.0)

    def test_pickle_VarArray(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.x[1]), None)
        model.junk.set_value(model.x, 1.0)
        self.assertEqual(model.junk.get(model.x), None)
        self.assertEqual(model.junk.get(model.x[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.x[1]), None)
        self.assertEqual(inst.junk.get(inst.x[1]), 1.0)

    def test_pickle_VarData(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.x[1]), None)
        model.junk.set_value(model.x[1], 1.0)
        self.assertEqual(model.junk.get(model.x[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.x[1]), None)
        self.assertEqual(inst.junk.get(inst.x[1]), 1.0)

    def test_pickle_ConstraintElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=model.x == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.c), None)
        model.junk.set_value(model.c, 1.0)
        self.assertEqual(model.junk.get(model.c), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.c), None)
        self.assertEqual(inst.junk.get(inst.c), 1.0)

    def test_pickle_ConstraintArray(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.c = Constraint([1, 2, 3], rule=simple_con_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.c), None)
        self.assertEqual(model.junk.get(model.c[1]), None)
        model.junk.set_value(model.c, 1.0)
        self.assertEqual(model.junk.get(model.c), None)
        self.assertEqual(model.junk.get(model.c[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.c[1]), None)
        self.assertEqual(inst.junk.get(inst.c[1]), 1.0)

    def test_pickle_ConstraintData(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.c = Constraint([1, 2, 3], rule=simple_con_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.c[1]), None)
        model.junk.set_value(model.c[1], 1.0)
        self.assertEqual(model.junk.get(model.c[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.c[1]), None)
        self.assertEqual(inst.junk.get(inst.c[1]), 1.0)

    def test_pickle_ObjectiveElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.obj = Objective(expr=model.x)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.obj), None)
        model.junk.set_value(model.obj, 1.0)
        self.assertEqual(model.junk.get(model.obj), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.obj), None)
        self.assertEqual(inst.junk.get(inst.obj), 1.0)

    def test_pickle_ObjectiveArray(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.obj = Objective([1, 2, 3], rule=simple_obj_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.obj), None)
        self.assertEqual(model.junk.get(model.obj[1]), None)
        model.junk.set_value(model.obj, 1.0)
        self.assertEqual(model.junk.get(model.obj), None)
        self.assertEqual(model.junk.get(model.obj[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.obj[1]), None)
        self.assertEqual(inst.junk.get(inst.obj[1]), 1.0)

    def test_pickle_ObjectiveData(self):
        model = ConcreteModel()
        model.x = Var([1, 2, 3], dense=True)
        model.obj = Objective([1, 2, 3], rule=simple_obj_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.obj[1]), None)
        model.junk.set_value(model.obj[1], 1.0)
        self.assertEqual(model.junk.get(model.obj[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.obj[1]), None)
        self.assertEqual(inst.junk.get(inst.obj[1]), 1.0)

    def test_pickle_SimpleBlock(self):
        model = ConcreteModel()
        model.b = Block()
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.b), None)
        model.junk.set_value(model.b, 1.0)
        self.assertEqual(model.junk.get(model.b), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.b), None)
        self.assertEqual(inst.junk.get(inst.b), 1.0)

    def test_pickle_IndexedBlock(self):
        model = ConcreteModel()
        model.b = Block([1, 2, 3])
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.b), None)
        self.assertEqual(model.junk.get(model.b[1]), None)
        model.junk.set_value(model.b, 1.0)
        self.assertEqual(model.junk.get(model.b), None)
        self.assertEqual(model.junk.get(model.b[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.b[1]), None)
        self.assertEqual(inst.junk.get(inst.b[1]), 1.0)

    def test_pickle_BlockData(self):
        model = ConcreteModel()
        model.b = Block([1, 2, 3])
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model.b[1]), None)
        model.junk.set_value(model.b[1], 1.0)
        self.assertEqual(model.junk.get(model.b[1]), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model.b[1]), None)
        self.assertEqual(inst.junk.get(inst.b[1]), 1.0)

    def test_pickle_model(self):
        model = ConcreteModel()
        model.junk = Suffix()
        self.assertEqual(model.junk.get(model), None)
        model.junk.set_value(model, 1.0)
        self.assertEqual(model.junk.get(model), 1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.get(model), None)
        self.assertEqual(inst.junk.get(inst), 1.0)


class TestSuffixFinder(unittest.TestCase):
    def test_suffix_finder(self):
        # Build a dummy model
        m = ConcreteModel()
        m.v1 = Var()

        m.b1 = Block()
        m.b1.v2 = Var()

        m.b1.b2 = Block()
        m.b1.b2.v3 = Var([0])

        # Add Suffixes
        m.suffix = Suffix(direction=Suffix.EXPORT)
        # No suffix on b1 - make sure we can handle missing suffixes
        m.b1.b2.suffix = Suffix(direction=Suffix.EXPORT)

        _suffix_finder = SuffixFinder('suffix')
        _suffix_b1_finder = SuffixFinder('suffix', context=m.b1)
        _suffix_b2_finder = SuffixFinder('suffix', context=m.b1.b2)

        # Check for no suffix value
        self.assertEqual(_suffix_finder.find(m.b1.b2.v3[0]), None)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2.v3[0]), None)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2.v3[0]), None)

        # Check finding default values
        # Add a default at the top level
        m.suffix[None] = 1
        self.assertEqual(_suffix_finder.find(m.b1.b2.v3[0]), 1)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2.v3[0]), None)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2.v3[0]), None)

        # Add a default suffix at a lower level
        m.b1.b2.suffix[None] = 2
        self.assertEqual(_suffix_finder.find(m.b1.b2.v3[0]), 2)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2.v3[0]), 2)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2.v3[0]), 2)

        # Check for container at lowest level
        m.b1.b2.suffix[m.b1.b2.v3] = 3
        self.assertEqual(_suffix_finder.find(m.b1.b2.v3[0]), 3)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2.v3[0]), 3)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2.v3[0]), 3)

        # Check for container at top level
        m.suffix[m.b1.b2.v3] = 4
        self.assertEqual(_suffix_finder.find(m.b1.b2.v3[0]), 4)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2.v3[0]), 3)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2.v3[0]), 3)

        # Check for specific values at lowest level
        m.b1.b2.suffix[m.b1.b2.v3[0]] = 5
        self.assertEqual(_suffix_finder.find(m.b1.b2.v3[0]), 5)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2.v3[0]), 5)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2.v3[0]), 5)

        # Check for specific values at top level
        m.suffix[m.b1.b2.v3[0]] = 6
        self.assertEqual(_suffix_finder.find(m.b1.b2.v3[0]), 6)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2.v3[0]), 5)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2.v3[0]), 5)

        # Make sure we don't find default suffixes at lower levels
        self.assertEqual(_suffix_finder.find(m.b1.v2), 1)
        self.assertEqual(_suffix_b1_finder.find(m.b1.v2), None)
        # Except if the lower level is actually the context
        self.assertEqual(_suffix_b2_finder.find(m.b1.v2), 2)

        # Make sure we don't find specific suffixes at lower levels
        m.b1.b2.suffix[m.v1] = 5
        self.assertEqual(_suffix_finder.find(m.v1), 1)
        self.assertEqual(_suffix_b1_finder.find(m.v1), None)
        # Except if the lower level is actually the context
        self.assertEqual(_suffix_b2_finder.find(m.v1), 5)

        # Make sure we can look up Blocks and that they will match
        # suffixes that they hold
        self.assertEqual(_suffix_finder.find(m.b1.b2), 2)
        self.assertEqual(_suffix_b1_finder.find(m.b1.b2), 2)
        self.assertEqual(_suffix_b2_finder.find(m.b1.b2), 2)

        self.assertEqual(_suffix_finder.find(m.b1), 1)
        self.assertEqual(_suffix_b1_finder.find(m.b1), None)
        self.assertEqual(_suffix_b2_finder.find(m.b1), 2)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.numvalue import (
    value,
    is_constant,
    is_fixed,
    is_variable_type,
    is_potentially_variable,
    NumericValue,
    ZeroConstant,
    native_numeric_types,
    native_types,
    polynomial_degree,
)

from pyomo.core.expr.boolean_value import BooleanValue

from pyomo.core.expr import (
    linear_expression,
    nonlinear_expression,
    land,
    lor,
    equivalent,
    exactly,
    atleast,
    atmost,
    all_different,
    count_if,
    implies,
    lnot,
    xor,
    inequality,
    log,
    log10,
    sin,
    cos,
    tan,
    cosh,
    sinh,
    tanh,
    asin,
    acos,
    atan,
    exp,
    sqrt,
    asinh,
    acosh,
    atanh,
    ceil,
    floor,
    Expr_if,
)

from pyomo.core.expr.calculus.derivatives import differentiate
from pyomo.core.expr.taylor_series import taylor_series_expansion

from pyomo.common.collections import ComponentMap
from pyomo.core.expr.symbol_map import SymbolMap
from pyomo.core.expr import (
    numvalue,
    numeric_expr,
    boolean_value,
    logical_expr,
    symbol_map,
    sympy_tools,
    taylor_series,
    visitor,
    expr_common,
    expr_errors,
    calculus,
)

from pyomo.core import expr, util, kernel

from pyomo.core.expr.numvalue import (
    nonpyomo_leaf_types,
    PyomoObject,
    native_numeric_types,
    value,
    is_constant,
    is_fixed,
    is_variable_type,
    is_potentially_variable,
    polynomial_degree,
    NumericValue,
    ZeroConstant,
)
from pyomo.core.expr.boolean_value import (
    as_boolean,
    BooleanConstant,
    BooleanValue,
    native_logical_values,
)
from pyomo.core.base import minimize, maximize
from pyomo.core.base.config import PyomoOptions

from pyomo.core.base.expression import Expression
from pyomo.core.base.label import (
    CuidLabeler,
    CounterLabeler,
    NumericLabeler,
    CNameLabeler,
    TextLabeler,
    AlphaNumericTextLabeler,
    NameLabeler,
    ShortNameLabeler,
)

#
# Components
#
from pyomo.core.base.component import name, Component, ModelComponentFactory
from pyomo.core.base.componentuid import ComponentUID
from pyomo.core.base.action import BuildAction
from pyomo.core.base.check import BuildCheck
from pyomo.core.base.set import Set, SetOf, simple_set_rule, RangeSet
from pyomo.core.base.param import Param
from pyomo.core.base.var import Var, ScalarVar, VarList
from pyomo.core.base.boolean_var import BooleanVar, BooleanVarList, ScalarBooleanVar
from pyomo.core.base.constraint import (
    simple_constraint_rule,
    simple_constraintlist_rule,
    ConstraintList,
    Constraint,
)
from pyomo.core.base.logical_constraint import LogicalConstraint, LogicalConstraintList
from pyomo.core.base.objective import (
    simple_objective_rule,
    simple_objectivelist_rule,
    Objective,
    ObjectiveList,
)
from pyomo.core.base.connector import Connector
from pyomo.core.base.sos import SOSConstraint
from pyomo.core.base.piecewise import Piecewise
from pyomo.core.base.suffix import (
    active_export_suffix_generator,
    active_import_suffix_generator,
    Suffix,
)
from pyomo.core.base.external import ExternalFunction
from pyomo.core.base.symbol_map import symbol_map_from_instance
from pyomo.core.base.reference import Reference

from pyomo.core.base.set import (
    Reals,
    PositiveReals,
    NonPositiveReals,
    NegativeReals,
    NonNegativeReals,
    Integers,
    PositiveIntegers,
    NonPositiveIntegers,
    NegativeIntegers,
    NonNegativeIntegers,
    Boolean,
    Binary,
    Any,
    AnyWithNone,
    EmptySet,
    UnitInterval,
    PercentFraction,
    RealInterval,
    IntegerInterval,
)
from pyomo.core.base.misc import display
from pyomo.core.base.block import (
    SortComponents,
    TraversalStrategy,
    Block,
    ScalarBlock,
    active_components,
    components,
    active_components_data,
    components_data,
)
from pyomo.core.base.PyomoModel import (
    global_option,
    Model,
    ConcreteModel,
    AbstractModel,
)
from pyomo.core.base.transformation import (
    Transformation,
    TransformationFactory,
    ReverseTransformationToken,
)

from pyomo.core.base.instance2dat import instance2dat

from pyomo.core.util import (
    prod,
    quicksum,
    sum_product,
    dot_product,
    summation,
    sequence,
)

# These APIs are deprecated and should be removed in the near future
from pyomo.core.base.set import set_options, RealSet, IntegerSet, BooleanSet

from pyomo.common.deprecation import relocated_module_attribute

relocated_module_attribute(
    'SimpleBlock', 'pyomo.core.base.block.SimpleBlock', version='6.0'
)
relocated_module_attribute('SimpleVar', 'pyomo.core.base.var.SimpleVar', version='6.0')
relocated_module_attribute(
    'SimpleBooleanVar', 'pyomo.core.base.boolean_var.SimpleBooleanVar', version='6.0'
)
del relocated_module_attribute



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy as numpy, numpy_available

import pyomo.environ as pyo
from pyomo.common import unittest
import pyomo.opt

import pyomo.contrib.alternative_solutions.tests.test_cases as tc
from pyomo.contrib.alternative_solutions import lp_enum

#
# Find available solvers. Just use GLPK if it's available.
#
solvers = list(
    pyomo.opt.check_available_solvers("glpk", "gurobi")
)  # , "appsi_gurobi"))
pytestmark = unittest.pytest.mark.parametrize("mip_solver", solvers)

timelimit = {"gurobi": "TimeLimit", "appsi_gurobi": "TimeLimit", "glpk": "tmlim"}


@unittest.pytest.mark.default
class TestLPEnum:

    def test_bad_solver(self, mip_solver):
        """
        Confirm that an exception is thrown with a bad solver name.
        """
        m = tc.get_3d_polyhedron_problem()
        try:
            lp_enum.enumerate_linear_solutions(m, solver="unknown_solver")
        except pyomo.common.errors.ApplicationError as e:
            pass

    @unittest.skipIf(True, "Ignoring fragile test for solver timeout.")
    def test_no_time(self, mip_solver):
        """
        Check that the correct bounds are found for a discrete problem where
        more restrictive bounds are implied by the constraints.
        """
        m = tc.get_3d_polyhedron_problem()
        with unittest.pytest.raises(Exception):
            lp_enum.enumerate_linear_solutions(
                m, solver=mip_solver, solver_options={timelimit[mip_solver]: 0}
            )

    def test_3d_polyhedron(self, mip_solver):
        m = tc.get_3d_polyhedron_problem()
        m.o.deactivate()
        m.obj = pyo.Objective(expr=m.x[0] + m.x[1] + m.x[2])

        sols = lp_enum.enumerate_linear_solutions(m, solver=mip_solver)
        assert len(sols) == 2
        for s in sols:
            assert s.objective_value == unittest.pytest.approx(4)

    def test_3d_polyhedron(self, mip_solver):
        m = tc.get_3d_polyhedron_problem()
        m.o.deactivate()
        m.obj = pyo.Objective(expr=m.x[0] + 2 * m.x[1] + 3 * m.x[2])

        sols = lp_enum.enumerate_linear_solutions(m, solver=mip_solver)
        assert len(sols) == 2
        for s in sols:
            assert s.objective_value == unittest.pytest.approx(
                9
            ) or s.objective_value == unittest.pytest.approx(10)

    def test_2d_diamond_problem(self, mip_solver):
        m = tc.get_2d_diamond_problem()
        sols = lp_enum.enumerate_linear_solutions(m, solver=mip_solver, num_solutions=2)
        assert len(sols) == 2
        for s in sols:
            print(s)
        assert sols[0].objective_value == unittest.pytest.approx(6.789473684210527)
        assert sols[1].objective_value == unittest.pytest.approx(3.6923076923076916)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_pentagonal_pyramid(self, mip_solver):
        n = tc.get_pentagonal_pyramid_mip()
        n.o.sense = pyo.minimize
        n.x.domain = pyo.Reals
        n.y.domain = pyo.Reals

        sols = lp_enum.enumerate_linear_solutions(n, solver=mip_solver, tee=False)
        for s in sols:
            print(s)
        assert len(sols) == 6

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_pentagon(self, mip_solver):
        n = tc.get_pentagonal_lp()

        sols = lp_enum.enumerate_linear_solutions(n, solver=mip_solver)
        for s in sols:
            print(s)
        assert len(sols) == 6



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

import pyomo.contrib.pynumero.interfaces.utils as utils


class TestCondensedSparseSummation(unittest.TestCase):
    def test_condensed_sparse_summation(self):
        data = [1.0, 0.0]
        row = [1, 2]
        col = [2, 2]
        A = scipy.sparse.coo_matrix((data, (row, col)), shape=(3, 3))

        data = [3.0, 0.0]
        B = scipy.sparse.coo_matrix((data, (row, col)), shape=(3, 3))

        # By default, scipy will remove structural nonzeros that
        # have zero values
        C = A + B
        self.assertEqual(C.nnz, 1)

        # Our CondensedSparseSummation should not remove any
        # structural nonzeros
        sparse_sum = utils.CondensedSparseSummation([A, B])
        C = sparse_sum.sum([A, B])
        expected_data = np.asarray([4.0, 0.0], dtype=np.float64)
        expected_row = np.asarray([1, 2], dtype=np.int64)
        expected_col = np.asarray([2, 2], dtype=np.int64)
        self.assertTrue(np.array_equal(expected_data, C.data))
        self.assertTrue(np.array_equal(expected_row, C.row))
        self.assertTrue(np.array_equal(expected_col, C.col))

        B.data[1] = 5.0
        C = sparse_sum.sum([A, B])
        expected_data = np.asarray([4.0, 5.0], dtype=np.float64)
        self.assertTrue(np.array_equal(expected_data, C.data))
        self.assertTrue(np.array_equal(expected_row, C.row))
        self.assertTrue(np.array_equal(expected_col, C.col))

        B.data[1] = 0.0
        C = sparse_sum.sum([A, B])
        expected_data = np.asarray([4.0, 0.0], dtype=np.float64)
        self.assertTrue(np.array_equal(expected_data, C.data))
        self.assertTrue(np.array_equal(expected_row, C.row))
        self.assertTrue(np.array_equal(expected_col, C.col))

    def test_repeated_row_col(self):
        data = [1.0, 0.0, 2.0]
        row = [1, 2, 1]
        col = [2, 2, 2]
        A = scipy.sparse.coo_matrix((data, (row, col)), shape=(3, 3))

        data = [3.0, 0.0]
        row = [1, 2]
        col = [2, 2]
        B = scipy.sparse.coo_matrix((data, (row, col)), shape=(3, 3))

        # Our CondensedSparseSummation should not remove any
        # structural nonzeros
        sparse_sum = utils.CondensedSparseSummation([A, B])
        C = sparse_sum.sum([A, B])
        expected_data = np.asarray([6.0, 0.0], dtype=np.float64)
        expected_row = np.asarray([1, 2], dtype=np.int64)
        expected_col = np.asarray([2, 2], dtype=np.int64)
        self.assertTrue(np.array_equal(expected_data, C.data))
        self.assertTrue(np.array_equal(expected_row, C.row))
        self.assertTrue(np.array_equal(expected_col, C.col))


if __name__ == '__main__':
    TestCondensedSparseSummation().test_condensed_sparse_summation()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# Remove one layer of .latex_printer
# import statement is now:
#   from pyomo.contrib.latex_printer import latex_printer
try:
    from pyomo.contrib.latex_printer.latex_printer import latex_printer
except:
    pass
    # in this case, the dependencies are not installed, nothing will work



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO

from pyomo.common.collections import Bunch
from pyomo.common.config import ConfigBlock
from pyomo.common.errors import DeveloperError
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.gdpopt.config_options import _add_common_configs
from pyomo.contrib.gdpopt.create_oa_subproblems import (
    add_util_block,
    add_disjunct_list,
    add_boolean_variable_lists,
    add_algebraic_variable_list,
)
from pyomo.contrib.gdpopt import __version__
from pyomo.contrib.gdpopt.util import (
    get_main_elapsed_time,
    lower_logger_level_to,
    solve_continuous_problem,
    time_code,
)
from pyomo.core.base import Objective, value, minimize, maximize
from pyomo.core.staleflag import StaleFlagManager
from pyomo.opt import SolverResults
from pyomo.opt import TerminationCondition as tc
from pyomo.util.model_size import build_model_size_report


class _GDPoptAlgorithm:
    CONFIG = ConfigBlock("GDPopt")
    _add_common_configs(CONFIG)

    def __init__(self, **kwds):
        """
        This is a common init method for all the GDPopt algorithms, so that we
        correctly set up the config arguments and initialize the generic parts
        of the algorithm state.
        """
        self.config = self.CONFIG(kwds.pop('options', {}), preserve_implicit=True)
        self.config.set_value(kwds)

        # We store bounds, timing info, iteration count, incumbent, and the
        # expression of the original (possibly nonlinear) objective function.
        self.LB = float('-inf')
        self.UB = float('inf')
        self.timing = Bunch()
        self.initialization_iteration = 0
        self.iteration = 0

        self.incumbent_boolean_soln = None
        self.incumbent_continuous_soln = None

        self.original_obj = None
        self._dummy_obj = None
        self.original_util_block = None

        self.log_formatter = (
            '{:>9}   {:>15}   {:>11.5f}   {:>11.5f}   {:>8.2%}   {:>7.2f}  {}'
        )

    # Support use as a context manager under current solver API
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass

    def available(self, exception_flag=True):
        """Solver is always available. Though subsolvers may not be, they will
        raise an error when the time comes.
        """
        return True

    def license_is_valid(self):
        return True

    def version(self):
        """Return a 3-tuple describing the solver version."""
        return __version__

    _metasolver = False

    def solve(self, model, **kwds):
        """Solve the model.

        Parameters
        ----------
        model : Block
            the Pyomo model or block to be solved

        """
        # I'm going to be nice for now and intercept with a more informative
        # error message than what the Config system would give.
        alg = kwds.pop('algorithm', None)
        if alg is None:
            alg = kwds.pop('strategy', None)
        if alg is not None:
            raise ValueError(
                "Changing the algorithm in the solve method "
                "is not supported for algorithm-specific "
                "GDPopt solvers. Either use "
                "SolverFactory('gdpopt') or instantiate a "
                "solver with the algorithm you want to use."
            )

        config = self.config(kwds.pop('options', {}), preserve_implicit=True)
        config.set_value(kwds)

        with lower_logger_level_to(config.logger, tee=config.tee):
            self._log_solver_intro_message(config)

            try:
                with time_code(self.timing, 'total', is_main_timer=True):
                    results = self._gather_problem_info_and_solve_non_gdps(
                        model, config
                    )
                    # If it wasn't disjunctive, we solved it
                    if not results:
                        # main loop implemented by each algorithm
                        self._solve_gdp(model, config)

            finally:
                self._get_final_pyomo_results_object()
                self._log_termination_message(config.logger)
                if self.algorithm == "LBB":
                    config.logger.warning(
                        "09/06/22: The GDPopt LBB algorithm currently has "
                        "known issues. Please use the results with caution "
                        "and report any bugs!"
                    )
                if self.pyomo_results.solver.termination_condition not in {
                    tc.infeasible,
                    tc.unbounded,
                }:
                    self._transfer_incumbent_to_original_model(config.logger)
                self._delete_original_model_util_block()
            return self.pyomo_results

    def _solve_gdp(self, original_model, config):
        # To be implemented by the algorithms
        raise NotImplementedError(
            "Derived _GDPoptAlgorithms need to implement the _solve_gdp method."
        )

    def _log_citation(self, config):
        pass

    def _log_solver_intro_message(self, config):
        config.logger.info(
            "Starting GDPopt version %s using %s algorithm"
            % (".".join(map(str, self.version())), self.algorithm)
        )
        os = StringIO()
        config.display(ostream=os)
        config.logger.info(os.getvalue())

        config.logger.info("""
            If you use this software, you may cite the following:
            - Implementation:
            Chen, Q; Johnson, ES; Bernal, DE; Valentin, R; Kale, S;
            Bates, J; Siirola, JD; Grossmann, IE.
            Pyomo.GDP: an ecosystem for logic based modeling and optimization
            development.
            Optimization and Engineering, 2021.
            """.strip())
        self._log_citation(config)

    def _log_header(self, logger):
        logger.info(
            '================================================================='
            '============================'
        )
        logger.info(
            '{:^9} | {:^15} | {:^11} | {:^11} | {:^8} | {:^7}\n'.format(
                'Iteration',
                'Subproblem Type',
                'Lower Bound',
                'Upper Bound',
                ' Gap ',
                'Time(s)',
            )
        )

    @property
    def objective_sense(self):
        if hasattr(self, 'pyomo_results'):
            return self.pyomo_results.problem.sense
        else:
            return None

    def _gather_problem_info_and_solve_non_gdps(self, model, config):
        """Solve the model.

        Args:
            model (Block): a Pyomo model or block to be solved

        """
        # set up the logger so that we will have a pretty log printed
        logger = config.logger

        self._create_pyomo_results_object_with_problem_info(model, config)
        # Check if this problem actually has any discrete decisions. If not,
        # just solve it.
        problem = self.pyomo_results.problem
        if (
            problem.number_of_binary_variables == 0
            and problem.number_of_integer_variables == 0
            and problem.number_of_disjunctions == 0
        ):
            cont_results = solve_continuous_problem(model, config)
            self.LB = cont_results.problem.lower_bound
            self.UB = cont_results.problem.upper_bound

            return self.pyomo_results

        # This class assumes that we have a util_block with an algebraic
        # variable list and a boolean variable list, so that we can transfer the
        # solution onto the original model. Everything else will be up to the
        # algorithm, but we make sure we have those here.

        # Make a block where we will store some component lists so that after we
        # clone we know who's who
        util_block = self.original_util_block = add_util_block(model)
        # Needed for finding indicator_vars mainly
        add_disjunct_list(util_block)
        add_boolean_variable_lists(util_block)
        # To transfer solutions between cloned models
        add_algebraic_variable_list(util_block)

    def _update_bounds_after_solve(
        self, subprob_nm, primal=None, dual=None, logger=None
    ):
        primal_improved = self._update_bounds(primal, dual)
        if logger is not None:
            self._log_current_state(logger, subprob_nm, primal_improved)

        return primal_improved

    def _update_bounds(self, primal=None, dual=None, force_update=False):
        """Update bounds correctly depending on objective sense.

        Args:
            primal: bound from solving subproblem with fixed discrete problem
            solution dual: bound from solving discrete problem (relaxation of
                           original problem)
            force_update: flag so this function will set the bound
                          even if it's not an improvement. (Used at termination
                          if the bounds cross.)
        """
        oldLB = self.LB
        oldUB = self.UB
        primal_bound_improved = False

        if self.objective_sense is minimize:
            if primal is not None and (primal < oldUB or force_update):
                self.UB = primal
                primal_bound_improved = primal < oldUB
            if dual is not None and (dual > oldLB or force_update):
                self.LB = dual
        else:
            if primal is not None and (primal > oldLB or force_update):
                self.LB = primal
                primal_bound_improved = primal > oldLB
            if dual is not None and (dual < oldUB or force_update):
                self.UB = dual

        return primal_bound_improved

    def relative_gap(self):
        """Returns current relative optimality gap.

        Note that this gap is not necessarily monotonically decreasing if at
        some point the primal bound changes signs.
        """
        absolute_gap = abs(self.UB - self.LB)
        return absolute_gap / (abs(self.primal_bound() + 1e-10))

    def _log_current_state(self, logger, subproblem_type, primal_improved=False):
        star = "*" if primal_improved else ""
        logger.info(
            self.log_formatter.format(
                self.iteration,
                subproblem_type,
                self.LB,
                self.UB,
                self.relative_gap(),
                get_main_elapsed_time(self.timing),
                star,
            )
        )

    def _log_termination_message(self, logger):
        logger.info(
            '\nSolved in {} iterations and {:.5f} seconds\n'
            'Optimal objective value {:.10f}\n'
            'Relative optimality gap {:.5%}'.format(
                self.iteration,
                get_main_elapsed_time(self.timing),
                self.primal_bound(),
                self.relative_gap(),
            )
        )

    def primal_bound(self):
        if self.objective_sense is minimize:
            return self.UB
        else:
            return self.LB

    def update_incumbent(self, util_block):
        self.incumbent_continuous_soln = [
            v.value for v in util_block.algebraic_variable_list
        ]
        self.incumbent_boolean_soln = [
            v.value for v in util_block.transformed_boolean_variable_list
        ]

    def _update_bounds_after_discrete_problem_solve(
        self, mip_termination, obj_expr, logger
    ):
        if mip_termination is tc.optimal:
            self._update_bounds_after_solve(
                'discrete', dual=value(obj_expr), logger=logger
            )
        elif mip_termination is tc.infeasible:
            # Discrete problem was infeasible.
            self._update_dual_bound_to_infeasible()
        elif mip_termination is tc.feasible or tc.unbounded:
            # we won't update the bound, because we didn't solve to
            # optimality. (And in the unbounded case, we wouldn't be here if we
            # didn't find a solution, so we're going to keep going, but we don't
            # have any info in terms of a dual bound.)
            pass
        else:
            raise DeveloperError(
                "Unrecognized termination condition %s when "
                "updating the dual bound." % mip_termination
            )

    def _update_dual_bound_to_infeasible(self):
        # set optimistic bound to infinity
        if self.objective_sense == minimize:
            self._update_bounds(dual=float('inf'))
        else:
            self._update_bounds(dual=float('-inf'))

    def _update_primal_bound_to_unbounded(self, config):
        if self.objective_sense == minimize:
            self._update_bounds(primal=float('-inf'))
        else:
            self._update_bounds(primal=float('inf'))
        config.logger.info('GDPopt exiting--GDP is unbounded.')
        self.pyomo_results.solver.termination_condition = tc.unbounded

    def _load_infeasible_termination_status(self, config):
        config.logger.info('GDPopt exiting--problem is infeasible.')
        self.pyomo_results.solver.termination_condition = tc.infeasible

    def bounds_converged(self, config):
        if self.pyomo_results.solver.termination_condition == tc.unbounded:
            return True
        elif self.LB + config.bound_tolerance >= self.UB:
            if self.LB == float('inf') and self.UB == float('inf'):
                self._load_infeasible_termination_status(config)
            elif self.LB == float('-inf') and self.UB == float('-inf'):
                self._load_infeasible_termination_status(config)
            else:
                # if they've crossed, then the gap is actually 0: Update the
                # dual (discrete problem) bound to be equal to the primal
                # (subproblem) bound
                if self.LB + config.bound_tolerance > self.UB:
                    self._update_bounds(dual=self.primal_bound(), force_update=True)
                self._log_current_state(config.logger, '')
                config.logger.info('GDPopt exiting--bounds have converged or crossed.')
                self.pyomo_results.solver.termination_condition = tc.optimal

            return True
        return False

    def reached_iteration_limit(self, config):
        if config.iterlim is not None and self.iteration >= config.iterlim:
            config.logger.info(
                'GDPopt unable to converge bounds within iteration limit of '
                '{} iterations.'.format(config.iterlim)
            )
            self.pyomo_results.solver.termination_condition = tc.maxIterations
            return True
        return False

    def reached_time_limit(self, config):
        elapsed = get_main_elapsed_time(self.timing)
        if config.time_limit is not None and elapsed >= config.time_limit:
            config.logger.info(
                'GDPopt exiting--Did not converge bounds '
                'before time limit of {} seconds. '.format(config.time_limit)
            )
            self.pyomo_results.solver.termination_condition = tc.maxTimeLimit
            return True
        return False

    def any_termination_criterion_met(self, config):
        return (
            self.bounds_converged(config)
            or self.reached_iteration_limit(config)
            or self.reached_time_limit(config)
        )

    def _create_pyomo_results_object_with_problem_info(self, original_model, config):
        """
        Initialize a results object with results.problem information
        """
        results = self.pyomo_results = SolverResults()

        results.solver.name = 'GDPopt %s - %s' % (self.version(), self.algorithm)

        prob = results.problem
        prob.name = original_model.name
        prob.number_of_nonzeros = None  # TODO

        num_of = build_model_size_report(original_model)

        # Get count of constraints and variables
        prob.number_of_constraints = num_of.activated.constraints
        prob.number_of_disjunctions = num_of.activated.disjunctions
        prob.number_of_variables = num_of.activated.variables
        prob.number_of_binary_variables = num_of.activated.binary_variables
        prob.number_of_continuous_variables = num_of.activated.continuous_variables
        prob.number_of_integer_variables = num_of.activated.integer_variables

        config.logger.info(
            "Original model has %s constraints (%s nonlinear) "
            "and %s disjunctions, "
            "with %s variables, of which %s are binary, %s are integer, "
            "and %s are continuous."
            % (
                num_of.activated.constraints,
                num_of.activated.nonlinear_constraints,
                num_of.activated.disjunctions,
                num_of.activated.variables,
                num_of.activated.binary_variables,
                num_of.activated.integer_variables,
                num_of.activated.continuous_variables,
            )
        )

        # Handle missing or multiple objectives, and get sense
        active_objectives = list(
            original_model.component_data_objects(
                ctype=Objective, active=True, descend_into=True
            )
        )
        number_of_objectives = len(active_objectives)
        if number_of_objectives == 0:
            config.logger.warning(
                'Model has no active objectives. Adding dummy objective.'
            )
            self._dummy_obj = discrete_obj = Objective(expr=1)
            original_model.add_component(
                unique_component_name(original_model, 'dummy_obj'), discrete_obj
            )
        elif number_of_objectives > 1:
            raise ValueError('Model has multiple active objectives.')
        else:
            discrete_obj = active_objectives[0]
        prob.sense = minimize if discrete_obj.sense == 1 else maximize

        return results

    def _transfer_incumbent_to_original_model(self, logger):
        StaleFlagManager.mark_all_as_stale(delayed=False)
        if self.incumbent_boolean_soln is None:
            assert self.incumbent_continuous_soln is None
            # we don't have a solution to transfer
            logger.info("No feasible solutions found.")
            return
        for var, soln in zip(
            self.original_util_block.algebraic_variable_list,
            self.incumbent_continuous_soln,
        ):
            var.set_value(soln, skip_validation=True)
        for var, soln in zip(
            self.original_util_block.boolean_variable_list, self.incumbent_boolean_soln
        ):
            if soln is None:
                var.set_value(soln, skip_validation=True)
            elif soln > 0.5:
                var.set_value(True)
            else:
                var.set_value(False)

        StaleFlagManager.mark_all_as_stale(delayed=True)

    def _delete_original_model_util_block(self):
        """For cleaning up after a solve--we want the original model to be
        untouched except for the solution being loaded"""
        blk = self.original_util_block
        if blk is not None:
            blk.parent_block().del_component(blk)
        # We just deleted the linearized objective if we had one, so restore the
        # prior one.
        if self.original_obj is not None:
            self.original_obj.activate()
        if self._dummy_obj is not None:
            self._dummy_obj.parent_block().del_component(self._dummy_obj)
            self._dummy_obj = None

    def _get_final_pyomo_results_object(self):
        """
        Fill in the results.solver information onto the results object
        """
        results = self.pyomo_results
        # Finalize results object
        results.problem.lower_bound = self.LB
        results.problem.upper_bound = self.UB
        results.solver.iterations = self.iteration
        results.solver.timing = self.timing
        results.solver.user_time = self.timing.total
        results.solver.wallclock_time = self.timing.total

        return results



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os

from .dependencies import ctypes, multiprocessing


def _as_bytes(val):
    """Helper function to coerce a string to a bytes() object"""
    if isinstance(val, bytes):
        return val
    elif val is not None:
        return val.encode('utf-8')
    return None


def _as_unicode(val):
    """Helper function to coerce a string to a unicode() object"""
    if isinstance(val, str):
        return val
    elif val is not None:
        return val.decode()
    return None


def _attempt_ctypes_cdll(name):
    """Load a CDLL library, returning bool indicating success"""
    try:
        dll = ctypes.CDLL(name)
        return True
    except OSError:
        return False


def _load_dll(name, timeout=10):
    """Load a DLL with a timeout

    On some platforms and some DLLs (notably Windows GitHub Actions with
    Python 3.5, 3.6, and 3.7 and the msvcr90.dll) we have observed
    behavior where the ctypes.CDLL() call hangs indefinitely. This uses
    multiprocessing to attempt the import in a subprocess (with a
    timeout) and then only calls the import in the main process if the
    subprocess succeeded.

    Performance note: CtypesEnviron only ever attempts to load a DLL
    once (the DLL reference is then held in a class attribute), and this
    interface only spawns the subprocess if ctypes.util.find_library
    actually locates the target library. This will have a measurable
    impact on Windows (where the DLLs exist), but not on other platforms.

    The default timeout of 10 is arbitrary. For simple situations, 1
    seems adequate. However, more complex examples have been observed
    that needed timeout==5. Using a default of 10 is simply doubling
    that observed case.

    """
    if not ctypes.util.find_library(name):
        return False, None

    if _load_dll.pool is None:
        # Resolving the deferred multiprocessing import could change the
        # local "multiprocessing" variable (replacing it with the
        # imported module).  This can result in an UnboundLocalError.
        # By explicitly declaring it "global" we can avoid the error.
        global multiprocessing
        try:
            _load_dll.pool = multiprocessing.Pool(1)
        except AssertionError:
            # multiprocessing will fail with an assertion error if this
            # Python process is a daemonic process (e.g., it was
            # launched within a dask server).  Fall back on a serial
            # process (and live with the risk that the import hangs).
            import multiprocessing.dummy

            _load_dll.pool = multiprocessing.dummy.Pool(1)
    job = _load_dll.pool.apply_async(_attempt_ctypes_cdll, (name,))
    try:
        result = job.get(timeout)
    except multiprocessing.TimeoutError:
        result = False
        # If there was a timeout, then the subprocess is likely hung and
        # cannot be reused.  Terminate it here (subsequent calls will
        # create a new Pool)
        _load_dll.pool.terminate()
        _load_dll.pool = None
    if result:
        return result, ctypes.CDLL(name)
    else:
        return result, None


# For efficiency, cache the multiprocessing Pool between calls to _load_dll
_load_dll.pool = None


class _RestorableEnvironInterface:
    """Interface to track environment changes and restore state"""

    def __init__(self, dll):
        assert dll.available()
        self.dll = dll
        self._original_state = {}

        # Transfer over the current os.environ
        for key, val in list(os.environ.items()):
            if val != self[key]:
                self[key] = val

        # If we can get a dictionary of the current environment (not
        # always possible), then remove any keys that are not in
        # os.environ
        origEnv = self.dll.get_env_dict()
        if origEnv is not None:
            for key in origEnv:
                if key not in os.environ:
                    del self[key]

    def restore(self):
        for key, val in self._original_state.items():
            if not val:
                if self[key] is not None:
                    del self[key]
            else:
                self[key] = val
        self._original_state = {}

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.dll.wgetenv(key)
        else:
            return self.dll.getenv(key)

    def __setitem__(self, key, val):
        if key not in self._original_state:
            self._original_state[key] = self[key]

        if isinstance(key, str):
            if isinstance(val, str):
                self.dll.wputenv_s(key, val)
            else:
                self.dll.wputenv_s(key, _as_unicode(val))
        elif isinstance(val, str):
            self.dll.wputenv_s(_as_unicode(key), val)
        else:
            self.dll.putenv_s(key, val)

    def __delitem__(self, key):
        if key not in self._original_state:
            self._original_state[key] = self[key]

        if isinstance(key, str):
            self.dll.wputenv_s(key, u'')
        else:
            self.dll.putenv_s(key, b'')


class _OSEnviron:
    """Helper class to proxy a "DLL-like" interface to os.environ"""

    _libname = 'os.environ'

    def available(self):
        return True

    def get_env_dict(self):
        return dict(os.environ)

    def getenv(self, key):
        # environb is not always present and depends on how the
        # interpreter was compiled.  Fall back on casting environ if
        # it is not available.
        try:
            return os.environb.get(key, None)
        except AttributeError:
            return _as_bytes(os.environ.get(_as_unicode(key), None))

    def wgetenv(self, key):
        # PY2 doesn't distinguish, and PY3's environ is nominally
        # unicode.  We will coerce the result to unicode to guarantee
        # the result type.
        return _as_unicode(os.environ.get(key, None))

    def putenv_s(self, key, val):
        # Win32 convention deletes environ entries when the string is empty
        if not val:
            if key in os.environ:
                del os.environ[key]
            return

        os.environb[key] = val

    def wputenv_s(self, key, val):
        # Win32 convention deletes environ entries when the string is empty
        if not val:
            if key in os.environ:
                del os.environ[key]
            return

        os.environ[key] = val


class _MsvcrtDLL:
    """Helper class to manage the interface with the MSVCRT runtime"""

    def __init__(self, name):
        self._libname = name
        if name is None:
            self._loaded = False
        else:
            self._loaded = None
        self.dll = None

    def available(self):
        if self._loaded is not None:
            return self._loaded

        self._loaded, self.dll = _load_dll(self._libname)
        if not self._loaded:
            return self._loaded

        self.putenv_s = self.dll._putenv_s
        self.putenv_s.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
        self.putenv_s.restype = ctypes.c_int

        self.wputenv_s = self.dll._wputenv_s
        self.wputenv_s.argtypes = [ctypes.c_wchar_p, ctypes.c_wchar_p]
        self.wputenv_s.restype = ctypes.c_int

        self.getenv = self.dll.getenv
        self.getenv.argtypes = [ctypes.c_char_p]
        self.getenv.restype = ctypes.c_char_p

        self.wgetenv = self.dll._wgetenv
        self.wgetenv.argtypes = [ctypes.c_wchar_p]
        self.wgetenv.restype = ctypes.c_wchar_p

        return self._loaded

    def get_env_dict(self):
        if not self.available():
            return None

        try:
            envp = ctypes.POINTER(ctypes.c_wchar_p).in_dll(self.dll, '_wenviron')
            if not envp.contents:
                envp = None
        except ValueError:
            envp = None
        if envp is None:
            try:
                envp = ctypes.POINTER(ctypes.c_char_p).in_dll(self.dll, '_environ')
                if not envp.contents:
                    return None
            except ValueError:
                return None

        ans = {}
        size = 0
        for line in envp:
            if not line:
                break
            size += len(line)
            if len(line) == 0:
                raise ValueError(
                    "Error processing MSVCRT _environ: 0-length string encountered"
                )
            if size > 32767:
                raise ValueError(
                    "Error processing MSVCRT _environ: "
                    "exceeded max environment block size (32767)"
                )
            key, val = line.split('=', 1)
            ans[key] = val
        return ans


class _Win32DLL:
    """Helper class to manage the interface with the Win32 runtime"""

    def __init__(self, name):
        self._libname = name
        if name is None:
            self._loaded = False
        else:
            self._loaded = None
        self.dll = None

    def available(self):
        if self._loaded is not None:
            return self._loaded

        self._loaded, self.dll = _load_dll(self._libname)
        if not self._loaded:
            return self._loaded

        self.putenv_s = self.dll.SetEnvironmentVariableA
        self.putenv_s.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
        self.putenv_s.restype = ctypes.c_bool

        self.wputenv_s = self.dll.SetEnvironmentVariableW
        self.wputenv_s.argtypes = [ctypes.c_wchar_p, ctypes.c_wchar_p]
        self.wputenv_s.restype = ctypes.c_bool

        # Note DWORD == c_ulong
        self._getenv_dll = self.dll.GetEnvironmentVariableA
        self._getenv_dll.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_ulong]
        self._getenv_dll.restype = ctypes.c_ulong

        self._wgetenv_dll = self.dll.GetEnvironmentVariableW
        self._wgetenv_dll.argtypes = [
            ctypes.c_wchar_p,
            ctypes.c_wchar_p,
            ctypes.c_ulong,
        ]
        self._wgetenv_dll.restype = ctypes.c_ulong

        # We (arbitrarily) choose to return the unicode environ
        self._envstr = self.dll.GetEnvironmentStringsW
        self._envstr.argtypes = []
        self._envstr.restype = ctypes.POINTER(ctypes.c_wchar)
        self._free_envstr = self.dll.FreeEnvironmentStringsW
        self._free_envstr.argtypes = [ctypes.POINTER(ctypes.c_wchar)]
        self._free_envstr.restype = ctypes.c_bool

        return self._loaded

    def getenv(self, key):
        size = self._getenv_dll(key, None, 0)
        if not size:
            return None
        buf = ctypes.create_string_buffer(b'\0' * size)
        self._getenv_dll(key, buf, size)
        return buf.value or None

    def wgetenv(self, key):
        size = self._wgetenv_dll(key, None, 0)
        if not size:
            return None
        buf = ctypes.create_unicode_buffer(u'\0' * size)
        self._wgetenv_dll(key, buf, size)
        return buf.value or None

    def get_env_dict(self):
        ans = {}
        _str_buf = self._envstr()
        # I am sure there is an easier way to parse this.
        # GetEnvironmentStringsW returns a single const char* that
        # points to a block of memory that contains the environment
        # strings.  Each environment string is NULL terminated, and an
        # empty string (effectively two consecutive NULLs) indicates the
        # end of the block.  The following just parses that data
        # character by character to reconstitute the original
        # environment strings.
        _null = {u'\0', b'\0'}
        i = 0
        while _str_buf[i] not in _null:
            _str = ''
            while _str_buf[i] not in _null:
                _str += _str_buf[i]
                i += len(_str_buf[i])
                if len(_str_buf[i]) == 0:
                    raise ValueError(
                        "Error processing Win32 GetEnvironmentStringsW: "
                        "0-length character encountered"
                    )
                if i > 32767:  # max var length
                    raise ValueError(
                        "Error processing Win32 GetEnvironmentStringsW: "
                        "exceeded max environment block size (32767)"
                    )
            key, val = _str.split('=', 1)
            ans[key] = val
            i += len(_str_buf[i])  # Skip the NULL
        self._free_envstr(_str_buf)
        return ans


class CtypesEnviron:
    """A context manager for managing environment variables

    This class provides a simplified interface for consistently setting
    and restoring environment variables, with special handling to ensure
    consistency with the C Runtime Library environment on Windows
    platforms.

    `os.environ` reflects the current python environment variables, and
    will be passed to subprocesses.  However, it does not reflect the C
    Runtime Library (MSVCRT) environment on Windows platforms.  This can
    be problemmatic as DLLs loaded through the CTYPES interface will see
    the MSVCRT environment and not os.environ.  This class provides a
    way to manage environment variables and pass changes to both
    os.environ and the MSVCRT runtime.

    This class implements a context manager API, so that clients can
    temporarily change - and then subsequently restore - the
    environment.

    .. testcode::
       :hide:

       import os
       orig_env_val = os.environ.get('TEMP_ENV_VAR', None)

    .. doctest::

       >>> os.environ['TEMP_ENV_VAR'] = 'original value'
       >>> print(os.environ['TEMP_ENV_VAR'])
       original value

       >>> with CtypesEnviron(TEMP_ENV_VAR='temporary value'):
       ...    print(os.environ['TEMP_ENV_VAR'])
       temporary value

       >>> print(os.environ['TEMP_ENV_VAR'])
       original value

    .. testcode::
       :hide:

       if orig_env_val is None:
           del os.environ['TEMP_ENV_VAR']
       else:
           os.environ['TEMP_ENV_VAR'] = orig_env_val

    """

    # Windows has a number of C Runtime Libraries, each of which can
    # hold its own independent copy of the system environment
    #
    # Note that this order is important: kernel32 appears to track
    # os.envion (and it is possible that it could be omitted).  It is
    # important to deal with it before the msvcrt libraries.
    DLLs = [
        _Win32DLL('kernel32'),
        _MsvcrtDLL(getattr(ctypes.util, 'find_msvcrt', lambda: None)()),
        _MsvcrtDLL('api-ms-win-crt-environment-l1-1-0'),
        _MsvcrtDLL('msvcrt'),
        _MsvcrtDLL('msvcr120'),
        _MsvcrtDLL('msvcr110'),
        _MsvcrtDLL('msvcr100'),
        _MsvcrtDLL('msvcr90'),
        _MsvcrtDLL('msvcr80'),
        _MsvcrtDLL('msvcr71'),
        _MsvcrtDLL('msvcr70'),
    ]

    def __init__(self, **kwds):
        self.interfaces = [_RestorableEnvironInterface(_OSEnviron())]
        self.interfaces.extend(
            _RestorableEnvironInterface(dll) for dll in self.DLLs if dll.available()
        )
        # If this is the first time a CtypesEnviron was created, the
        # calls to dll.activate() may have spawned a multiprocessing
        # pool, which we should clean up.
        if _load_dll.pool is not None:
            _load_dll.pool.terminate()
            _load_dll.pool = None
        # Set the incoming env strings on all interfaces...
        for k, v in kwds.items():
            self[k] = v

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.restore()

    def restore(self):
        """Restore the environment to the original state

        This restores all environment variables modified through this
        object to the state they were in before this instance made any
        changes.  Note that any changes made directly to os.environ
        outside this instance will not be detected / undone.

        """
        # It is possible that os.environ and the MSVCRT did not start
        # off in sync; e.g., if someone had set a value in os.environ
        # directly.  We will be especially careful and restore each
        # environ back to its original state.  Note that there are cases
        # where some of the libs refer to the same DLL space, so we will
        # restore them in the opposite order that we set them, so that
        # cases where multiple libs point to the same environment things
        # are all restored correctly.
        for lib in reversed(self.interfaces):
            lib.restore()

    def __getitem__(self, key):
        """Return the current environment variable value from os.environ"""
        return os.environ[key]

    def __contains__(self, key):
        """Return True if the key is in os.environ"""
        return key in os.environ

    def __setitem__(self, key, val):
        """Set an environment variable in all known runtime environments"""
        for lib in self.interfaces:
            lib[key] = val

    def __delitem__(self, key):
        """Remove an environment variable from all known runtime environments"""
        for lib in self.interfaces:
            del lib[key]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Collection of GDP-related hacks.

Hacks for dealing with the fact that solver writers may sometimes fail to
detect variables inside of Disjuncts or deactivated Blocks.
"""

import logging
from pyomo.core.base import Transformation, Block, Constraint
from pyomo.gdp import Disjunct, GDP_Error, Disjunction
from pyomo.core import TraversalStrategy, TransformationFactory
from pyomo.core.base.indexed_component import ActiveIndexedComponent
from pyomo.common.deprecation import deprecated

logger = logging.getLogger('pyomo.gdp')


@TransformationFactory.register('gdp.reclassify', doc="Reclassify Disjuncts to Blocks.")
class HACK_GDP_Disjunct_Reclassifier(Transformation):
    """Reclassify Disjuncts to Blocks.

    HACK: this will reclassify all Disjuncts to Blocks so the current writers
    can find the variables

    """

    @deprecated(
        msg="The gdp.reclasify transformation has been deprecated in "
        "favor of the gdp transformations creating References to "
        "variables local to each Disjunct during the transformation. "
        "Validation that the model has been completely transformed "
        "to an algebraic model has been moved to the "
        "check_model_algebraic function in gdp.util.",
        version='5.7',
    )
    def _apply_to(self, instance, **kwds):
        assert not kwds  # no keywords expected to the transformation
        disjunct_generator = instance.component_objects(
            Disjunct,
            descend_into=(Block, Disjunct),
            descent_order=TraversalStrategy.PostfixDFS,
        )
        for disjunct_component in disjunct_generator:
            # Check that the disjuncts being reclassified are all relaxed or
            # are not on an active block.
            for disjunct in disjunct_component.values():
                if (
                    disjunct.active
                    and self._disjunct_not_relaxed(disjunct)
                    and self._disjunct_on_active_block(disjunct)
                    and self._disjunct_not_fixed_true(disjunct)
                ):
                    # First, do a couple checks in order to give a more
                    # useful error message
                    disjunction_set = {
                        i
                        for i in instance.component_data_objects(
                            Disjunction, descend_into=True, active=None
                        )
                    }
                    active_disjunction_set = {
                        i
                        for i in instance.component_data_objects(
                            Disjunction, descend_into=True, active=True
                        )
                    }
                    disjuncts_in_disjunctions = set()
                    for i in disjunction_set:
                        disjuncts_in_disjunctions.update(i.disjuncts)
                    disjuncts_in_active_disjunctions = set()
                    for i in active_disjunction_set:
                        disjuncts_in_active_disjunctions.update(i.disjuncts)

                    if disjunct not in disjuncts_in_disjunctions:
                        raise GDP_Error(
                            'Disjunct "%s" is currently active, '
                            'but was not found in any Disjunctions. '
                            'This is generally an error as the model '
                            'has not been fully relaxed to a '
                            'pure algebraic form.' % (disjunct.name,)
                        )
                    elif disjunct not in disjuncts_in_active_disjunctions:
                        raise GDP_Error(
                            'Disjunct "%s" is currently active. While '
                            'it participates in a Disjunction, '
                            'that Disjunction is currently deactivated. '
                            'This is generally an error as the '
                            'model has not been fully relaxed to a pure '
                            'algebraic form. Did you deactivate '
                            'the Disjunction without addressing the '
                            'individual Disjuncts?' % (disjunct.name,)
                        )
                    else:
                        raise GDP_Error("""
                        Reclassifying active Disjunct "%s" as a Block.  This
                        is generally an error as it indicates that the model
                        was not completely relaxed before applying the
                        gdp.reclassify transformation""" % (disjunct.name,))

            # Reclassify this disjunct as a block
            disjunct_component.parent_block().reclassify_component_type(
                disjunct_component, Block
            )
            # HACK: activate the block, but do not activate the
            # BlockData objects
            super(ActiveIndexedComponent, disjunct_component).activate()

            # Deactivate all constraints.  Note that we only need to
            # descend into blocks: we will catch disjuncts in the outer
            # loop.
            #
            # Note that we defer this until AFTER we reactivate the
            # block, as the component_objects generator will not
            # return anything when active=True and the block is
            # deactivated.
            for disjunct in disjunct_component._data.values():
                if self._disjunct_not_relaxed(disjunct):
                    disjunct._deactivate_without_fixing_indicator()
                else:
                    disjunct._activate_without_unfixing_indicator()

                cons_in_disjunct = disjunct.component_objects(
                    Constraint, descend_into=Block, active=True
                )
                for con in cons_in_disjunct:
                    con.deactivate()

    def _disjunct_not_fixed_true(self, disjunct):
        # Return true if the disjunct indicator variable is not fixed to True
        return not (disjunct.indicator_var.fixed and disjunct.indicator_var.value)

    def _disjunct_not_relaxed(self, disjunct):
        # Return True if the disjunct was not relaxed by a transformation.
        return disjunct.transformation_block is None

    def _disjunct_on_active_block(self, disjunct):
        # Check first to make sure that the disjunct is not a
        # descendent of an inactive Block or fixed and deactivated
        # Disjunct, before raising a warning.
        parent_block = disjunct.parent_block()
        while parent_block is not None:
            if parent_block.ctype is Block and not parent_block.active:
                return False
            elif (
                parent_block.ctype is Disjunct
                and not parent_block.active
                and parent_block.indicator_var.value == False
                and parent_block.indicator_var.fixed
            ):
                return False
            else:
                # Step up one level in the hierarchy
                parent_block = parent_block.parent_block()
                continue
        return True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
import os

import pyomo.common.unittest as unittest

import pyomo.scripting.pyomo_main as main
from pyomo.opt import check_available_solvers

currdir = os.path.dirname(os.path.abspath(__file__))
exdir = os.path.abspath(
    os.path.join(currdir, '..', '..', '..', '..', 'examples', 'pyomo', 'diet')
)

sqlite3_available = pyodbc_available = False
try:
    import sqlite3

    sqlite3_available = True
except ImportError:
    pass
try:
    import pyodbc

    pyodbc_available = True
    #
    # Temporarily deprecating pyodbc tests.
    # These tests are not reliably executing with Python 2.6 and 2.7,
    # due to apparent issues with unicode representation.
    #
    pyodbc_available = False
except ImportError:
    pass

solvers = None


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global solvers
        import pyomo.environ

        solvers = check_available_solvers('glpk')

    def run_pyomo(self, *args, **kwargs):
        """
        Run Pyomo with the given arguments. `args` should be a list with
        one argument token per string item.
        """
        if not 'glpk' in solvers:
            self.skipTest("GLPK is not installed")

        if isinstance(args, str):
            args = args.split()
        args = list(map(str, args))

        outputpath = kwargs.pop('outputpath', os.path.join(exdir, 'results.jsn'))
        args = [
            'solve',
            '--solver=glpk',
            '--results-format=json',
            '-c',
            '--logging=quiet',
            '--save-results',
            outputpath,
        ] + args

        old_path = os.getcwd()
        os.chdir(exdir)
        output = main.main(args)
        os.chdir(old_path)
        return outputpath

    def compare_json(self, file1, file2):
        with open(file1, 'r') as f1, open(file2, 'r') as f2:
            file1_contents = json.load(f1)
            file2_contents = json.load(f2)
        self.assertStructuredAlmostEqual(
            file2_contents, file1_contents, allow_second_superset=True
        )

    def test_pyomo_dat(self):
        results_file = self.run_pyomo(
            os.path.join(exdir, 'diet1.py'),
            os.path.join(exdir, 'diet.dat'),
            outputpath=os.path.join(currdir, 'pyomo_dat.jsn'),
        )
        baseline_file = os.path.join(currdir, 'baselines', 'diet1_pyomo_dat.jsn')
        self.compare_json(results_file, baseline_file)

    @unittest.skipUnless(pyodbc_available, "Requires PyODBC")
    def test_pyomo_mdb(self):
        results_file = self.run_pyomo(
            os.path.join(exdir, 'diet1.py'),
            os.path.join(exdir, 'diet1.db.dat'),
            outputpath=os.path.join(currdir, 'pyomo_mdb.jsn'),
        )
        baseline_file = os.path.join(currdir, 'baselines', 'diet1_pyomo_mdb.jsn')
        self.compare_json(results_file, baseline_file)

    @unittest.skipUnless(pyodbc_available, "Requires PyODBC")
    def test_mdb_equality(self):
        dat_results_file = self.run_pyomo(
            os.path.join(exdir, 'diet1.py'),
            os.path.join(exdir, 'diet.dat'),
            outputpath=os.path.join(currdir, 'dat_results.jsn'),
        )
        with open(dat_results_file) as FILE:
            dat_results = json.load(FILE)
        db_results_file = self.run_pyomo(
            os.path.join(exdir, 'diet1.py'),
            os.path.join(exdir, 'diet1.db.dat'),
            outputpath=os.path.join(currdir, 'db_results.jsn'),
        )
        with open(db_results_file) as FILE:
            db_results = json.load(FILE)
        # Filter out the solver time
        del dat_results['Solver'][0]['Time']
        del db_results['Solver'][0]['Time']
        # Compare baselines
        self.assertStructuredAlmostEqual(dat_results, db_results)
        os.remove(dat_results_file)
        os.remove(db_results_file)

    @unittest.skipUnless(sqlite3_available, "Requires SQLite3")
    def test_pyomo_sqlite3(self):
        results_file = self.run_pyomo(
            os.path.join(exdir, 'diet1.py'),
            os.path.join(exdir, 'diet1.sqlite.dat'),
            outputpath=os.path.join(currdir, 'pyomo_sqlite3.jsn'),
        )
        baseline_file = os.path.join(currdir, 'baselines', 'diet1_pyomo_sqlite3.jsn')
        self.compare_json(results_file, baseline_file)

    @unittest.skipUnless(sqlite3_available, "Requires SQLite3")
    def test_sqlite_equality(self):
        dat_results_file = self.run_pyomo(
            os.path.join(exdir, 'diet1.py'),
            os.path.join(exdir, 'diet.dat'),
            outputpath=os.path.join(currdir, 'dat_results.jsn'),
        )
        with open(dat_results_file) as FILE:
            dat_results = json.load(FILE)
        sqlite_results_file = self.run_pyomo(
            os.path.join(exdir, 'diet1.py'),
            os.path.join(exdir, 'diet1.sqlite.dat'),
            outputpath=os.path.join(currdir, 'sqlite_results.jsn'),
        )
        with open(sqlite_results_file) as FILE:
            sqlite_results = json.load(FILE)
        # Filter out the solver time
        del dat_results['Solver'][0]['Time']
        del sqlite_results['Solver'][0]['Time']
        # Compare baselines
        self.assertStructuredAlmostEqual(dat_results, sqlite_results)
        os.remove(dat_results_file)
        os.remove(sqlite_results_file)


if __name__ == "__main__":
    unittest.main()
